           (*-------------------------------------------*
            |                Csp-Prover                 |
            |                July 2004                  |
            |        Yoshinao Isobe (AIST JAPAN)        |
            *-------------------------------------------*)

theory Csp_Hide = Csp_semantics:

(*  The following simplification rules are deleted in this theory file *)
(*  because they unexpectly rewrite (notick | t = []t)                 *)
(*                                                                     *)
(*                  disj_not1: (~ P | Q) = (P --> Q)                   *)

declare disj_not1 [simp del]

(*****************************************************************

         1. T --t X : dom_T
         2. F --f X : dom_F
         3. SF -- X : dom_SF
         4. 

 *****************************************************************)

(*********************************************************
                 Lemma (Hidet in Dom_T)
 *********************************************************)

lemma Hidet_in : 
  "ALL X S. S:dom_T --> S --t X: dom_T"
apply (intro allI impI)

apply (simp (no_asm) add: dom_T_def CT1_def)

  (* non-empty *)

apply (rule conjI)
apply (simp add: Hidet_def)
apply (rule_tac x="[]t" in exI)
apply (simp add: nil_t_in_all_dom)

  (* prefix closed *)

apply (simp add: prefix_closed_def)
apply (intro allI impI)
apply (simp add: Hidet_def)
apply (elim conjE exE)
apply (rename_tac X S v u s)
apply (simp)

apply (insert hide_trace_prefix)
apply (drule_tac x="X" in spec)
apply (drule_tac x="s" in spec)
apply (drule_tac x="v" in spec)
apply (simp)
apply (elim conjE exE)
apply (rule_tac x="t" in exI)
apply (simp)

apply (simp add: dom_T_def CT1_def)
apply (blast intro: prefix_closed_propI)
done

(*********************************************************
                 Lemma (Hidef in Dom_F)
 *********************************************************)

lemma Hidef_in : 
  "ALL X S. S:dom_F --> S --f X: dom_F"
apply (intro allI impI)

apply (simp (no_asm) add: dom_F_def CF1_def)
apply (intro allI impI)

apply (simp add: Hidef_def)
apply (elim conjE exE)
apply (rename_tac X S s Z Y t)
apply (rule_tac x="t" in exI)
apply (simp)

apply (simp add: dom_F_def)
apply (unfold CF1_def)

apply (drule_tac x="t" in spec)
apply (drule_tac x="Ev ` X Un Z" in spec)
apply (drule_tac x="Ev ` X Un Y" in spec)
apply (auto)
done

(*********************************************************
                 Lemma (Hidesf in Dom_SF)
 *********************************************************)

(*** T2 ***)

lemma Hidesf_in_T2 : "ALL X TF. TF:dom_SF --> CT2(TF --sf X)"
apply (intro allI impI)
apply (simp add: Hidesf_def CT2_def)
apply (intro allI impI)
apply (simp add: Hidet_in Rep_dom_T Abs_dom_T_inverse)
apply (simp add: Hidef_in Rep_dom_F Abs_dom_F_inverse)

apply (simp add: Hidet_def)
apply (elim conjE exE)

apply (insert hide_trace_reconst_tick)
apply (drule_tac x="X" in spec)
apply (drule_tac x="t" in spec)
apply (drule_tac x="s" in spec)
apply (simp)
apply (elim conjE exE)

apply (simp add: dom_SF_def CT2_def)
apply (erule conjE)+
apply (drule_tac x="t1" in spec)

apply (insert hide_trace_notick_inv)
apply (drule_tac x="X" in spec)
apply (drule_tac x="t1" in spec)
apply (simp)

apply (simp add: Hidef_def)
apply (rule conjI)
apply (rule_tac x="t1" in exI)
apply (simp add: Un_Evset)

apply (rule_tac x="t1 @t [Tick]t" in exI)
apply (simp add: image_def EvsetTick_def)
done

(*** F2 ***)

(* a condiction for F2 *)

lemma Hidesf_in_F2_notin :
  "ALL X TF s Y Z t. (notick s & s = t --tr X &
       (ALL a. a : Z --> s @t [a]t ~: Rep_dom_T (fst TF) --t X))
        --> (ALL a. a : (Z- Ev ` X) 
                     --> t @t [a]t ~: Rep_dom_T (fst TF))"

apply (intro allI impI)
apply (case_tac "t @t [a]t ~: Rep_dom_T (fst TF)")
apply (simp)

      (** t @t [a]t : Rep_dom_T (fst TF) --> contradiction **)

apply (simp)
apply (insert hide_trace_last_notin)
apply (drule_tac x="X" in spec)
apply (drule_tac x="t" in spec)
apply (drule_tac x="a" in spec)
apply (insert hide_trace_notick_inv)
apply (drule_tac x="X" in spec)
apply (drule_tac x="t" in spec)
apply (erule conjE)+
apply (simp)

apply (drule_tac x="a" in spec)
apply (simp add: Hidet_def)
apply (drule_tac x="t @t [a]t" in spec)
apply (simp)
done

lemma Hidesf_in_F2_set_cal : 
  "ALL X Y Z. Ev ` X Un Y Un (Z - Ev ` X) = Ev ` X Un (Y Un Z)"
apply (auto)
done

(* F2 *)

lemma Hidesf_in_F2 : "ALL X TF. TF:dom_SF --> CF2(TF --sf X)"
apply (intro allI impI)
apply (simp add: Hidesf_def CF2_def)
apply (intro allI impI)
apply (simp add: Hidet_in Rep_dom_T Abs_dom_T_inverse)
apply (simp add: Hidef_in Rep_dom_F Abs_dom_F_inverse)
apply (rename_tac X TF s Y Z)

apply (simp add: Hidef_def)
apply (elim conjE exE)
apply (rule_tac x="t" in exI)

apply (insert Hidesf_in_F2_notin)
apply (drule_tac x="X" in spec)
apply (drule_tac x="TF" in spec)
apply (drule_tac x="s" in spec)
apply (drule_tac x="Y" in spec)
apply (drule_tac x="Z" in spec)
apply (drule_tac x="t" in spec)
apply (simp)

apply (simp add: dom_SF_def CF2_def)
apply (erule conjE)+
apply (drule_tac x="t" in spec)
apply (drule_tac x="Ev ` X Un Y" in spec)
apply (drule_tac x="Z- Ev ` X" in spec)
apply (simp)

apply (insert hide_trace_notick_inv)
apply (drule_tac x="X" in spec)
apply (drule_tac x="t" in spec)
apply (simp add: Hidesf_in_F2_set_cal)
done

(*** F3 ***)

lemma Hidesf_in_F3 : "ALL X TF. TF:dom_SF --> CF3(TF --sf X)"
apply (intro allI impI)
apply (simp add: Hidesf_def CF3_def)
apply (intro allI impI)
apply (simp add: Hidet_in Rep_dom_T Abs_dom_T_inverse)
apply (simp add: Hidef_in Rep_dom_F Abs_dom_F_inverse)

apply (erule exE)
apply (rename_tac X TF s Y)

apply (simp add: Hidef_def)
apply (elim conjE exE)

apply (simp add: Hidet_def)
apply (rule_tac x="t" in exI)
apply (simp)

apply (simp add: dom_SF_def CF3_def)
apply (erule conjE)+
apply (drule_tac x="t" in spec)
apply (drule mp)
apply (rule_tac x="Ev ` X Un Y" in exI)
apply (simp_all)
done

(*** Hidesf in dom_SF ***)

lemma Hidesf_in :
  "ALL X TF. TF:dom_SF --> TF --sf X: dom_SF"
apply (intro allI impI)
apply (simp (no_asm) add: dom_SF_def)
apply (simp add: Hidesf_in_T2)
apply (simp add: Hidesf_in_F2)
apply (simp add: Hidesf_in_F3)
done

(*********************************************************
                     Lemmas (mono)
 *********************************************************)

(*** T ***)

lemma Hidet_mono: 
  "ALL X S T. S <= T --> S --t X <= T --t X"
apply (intro allI impI)
apply (simp add: Hidet_def)
apply (simp add: subset_def)
apply (intro allI impI)
apply (elim conjE exE)
apply (drule_tac x="t" in bspec)
apply (simp)
apply (rule_tac x="t" in exI)
apply (simp)
done

(*** F ***)

lemma Hidef_mono: 
  "ALL X S T. S <= T --> S --f X <= T --f X"
apply (intro allI impI)
apply (simp add: Hidef_def)
apply (simp add: subset_def)
apply (intro allI impI)
apply (elim conjE exE)
apply (drule_tac x="(t, Ev ` X Un b)" in bspec)
apply (simp)
apply (rule_tac x="t" in exI)
apply (simp)
done

lemma Hidesf_mono: 
  "ALL X SE TF. SE <= TF --> SE --sf X <= TF --sf X"
apply (auto simp add: Hidesf_def)
apply (simp add: order_pair_def)
apply (simp add: subset_T_def subset_F_def)
apply (simp add: Hidet_in Hidef_in Rep_dom_T Rep_dom_F
                 Abs_dom_T_inverse Abs_dom_F_inverse)
apply (simp add: Hidet_mono Hidef_mono)
done

(*********************************************************
                   Hide trace eval
 *********************************************************)

lemma Hidet_eval:
  "ALL X P ev. [[P -- X]]T ev = [[P]]T ev --t X"
apply (intro allI)
apply (simp add: eval_T_def)
apply (simp add: eval_TF_decompo)
apply (simp add: Hidesf_def)
apply (simp add: Hidet_in Rep_dom_T Abs_dom_T_inverse)
done

(*********************************************************
                   Hide failure eval
 *********************************************************)

lemma Hidef_eval:
  "ALL X P ev. [[P -- X]]F ev = [[P]]F ev --f X"
apply (intro allI)
apply (simp add: eval_F_def)
apply (simp add: eval_TF_decompo)
apply (simp add: Hidesf_def)
apply (simp add: Hidef_in Rep_dom_F Abs_dom_F_inverse)
done

(****************** to add it again ******************)

declare disj_not1 [simp] 

end
