           (*-------------------------------------------*
            |                Csp-Prover                 |
            |                July 2004                  |
            |        Yoshinao Isobe (AIST JAPAN)        |
            *-------------------------------------------*)

theory Csp_Sequence = Csp_semantics:

(*****************************************************************

         1. S ;t T     : dom_T
         2. <S,E> ;f F : dom_F
         3. SE ;sf TF  : dom_SF
         4. 

 *****************************************************************)

(*********************************************************
               Lemma (Sequencet in Dom_T)
 *********************************************************)

(*** lm 1 ***)

lemma Sequencet_in_lm1 : 
       "[| S : dom_T ; T : dom_T ; prefix s t ; t = rmtick sa ; sa : S |]
         ==> EX sa. s = rmtick sa & sa : S"
apply (insert rmtick_prefix)
apply (drule_tac x="s" in spec)
apply (drule_tac x="sa" in spec)
apply (simp)
apply (elim conjE exE)

apply (rule_tac x="u" in exI)
apply (simp)
apply (insert dom_T_is_prefix_closed)
apply (drule_tac x="u" in spec)
apply (drule_tac x="sa" in spec)
apply (drule_tac x="S" in spec)
apply (simp)
done

(*** lm 2 ***)

lemma Sequencet_in_lm2 : 
   "[| S : dom_T; T : dom_T; t = Abs_trace (Rep_trace s @ ua @ Rep_trace s2);
          notick s & Tick ~: set ua;
          Abs_trace (Rep_trace s @ ua @ Rep_trace [Tick]t) : S; s2 : T;
          Rep_trace s @ ua @ Rep_trace s2 : trace; Rep_trace s1 = Rep_trace s @ ua;
          u = ua @ Rep_trace s2 |]
       ==> EX sa. s = rmtick sa & sa : S"

apply (rule_tac x="s" in exI)
apply (simp add: rmtick_notick)

apply (insert dom_T_is_prefix_closed)
apply (drule_tac x="s" in spec)
apply (drule_tac x="Abs_trace (Rep_trace s @ ua @ Rep_trace [Tick]t)" in spec)
apply (drule_tac x="S" in spec)
apply (drule mp)

apply (simp add: prefix_list_def Rep_trace)
apply (rule_tac x="ua @ Rep_trace [Tick]t" in exI)

apply (simp add: event_in_trace Abs_trace_inverse)
apply (insert app_event_notick)
apply (drule_tac x="ua" in spec)
apply (drule_tac x="Tick" in spec)
apply (simp)

apply (insert app_trace_notick)
apply (drule_tac x="Rep_trace s" in spec)
apply (drule_tac x="ua @ [Tick]" in spec)
apply (simp add: Abs_trace_inverse)

apply (simp)
done

(*** main ***)

lemma Sequencet_in : 
  "ALL S T. S:dom_T & T:dom_T --> S ;;t T: dom_T"
apply (intro allI impI)

apply (simp (no_asm) add: dom_T_def CT1_def)

  (* non-empty *)

apply (rule conjI)
apply (simp add: Sequencet_def)
apply (simp add: dom_T_def CT1_def)

  (* prefix closed *)

apply (simp (no_asm) add: prefix_closed_def)
apply (intro allI impI)
apply (simp add: Sequencet_def)
apply (elim conjE exE)

apply (erule disjE)
apply (elim conjE exE)
apply (simp add: Sequencet_in_lm1)

apply (elim conjE exE)
apply (simp)

apply (simp add: prefix_list_def)
apply (elim conjE exE)

apply (simp add: trace_app_in_trace_if Abs_trace_inverse)
apply (insert list_app_app)
apply (drule_tac x="Rep_trace sa" in spec)
apply (drule_tac x="Rep_trace ta" in spec)
apply (drule_tac x="Rep_trace s" in spec)
apply (drule_tac x="u" in spec)
apply (simp)
apply (rename_tac S T s t s1 s2 u)

apply (erule disjE)
  (* Rep_trace s1 >= Rep_trace s *)

apply (elim conjE exE)
apply (rule disjI1)
apply (simp add: Sequencet_in_lm2)

  (* Rep_trace s1 <= Rep_trace s *)

apply (elim conjE exE)
apply (rule disjI2)
apply (simp)
apply (rule_tac x="s1" in exI)
apply (rule_tac x="Abs_trace ua" in exI)
apply (simp)

apply (insert app_trace_in_trace_inv)
apply (drule_tac x="ua" in spec)
apply (drule_tac x="u" in spec)
apply (rotate_tac -2)
apply (drule sym)
apply (simp add: Rep_trace)

apply (simp add: Abs_trace_inverse)
apply (simp add: Abs_Rep_t_inv_imp)

apply (insert dom_T_is_prefix_closed)
apply (drule_tac x="Abs_trace ua" in spec)
apply (drule_tac x="s2" in spec)
apply (drule_tac x="T" in spec)
apply (drule mp)

apply (simp add: prefix_list_def Rep_trace)
apply (rule_tac x="u" in exI)
apply (simp add: Abs_trace_inverse)
apply (simp)
done

(*********************************************************
               Lemma (Sequencef in Dom_F)
 *********************************************************)

lemma Sequencef_in : 
  "ALL T1 F1 F2. T1 : dom_T & F1 : dom_F & F2 : dom_F 
     --> (T1, F1) ;;f F2: dom_F"
apply (intro allI impI)

apply (simp (no_asm) add: dom_F_def CF1_def)
apply (intro allI impI)

apply (simp add: Sequencef_def)
apply (elim conjE)

apply (erule disjE)

 (* (s, insert Tick Y) : F1 *)

apply (rule disjI1)

apply (simp add: dom_F_def)
apply (unfold CF1_def)
apply (drule_tac x="s" in spec)
apply (drule_tac x="insert Tick X" in spec)
apply (drule_tac x="insert Tick Y" in spec)
apply (force)

 (* s = sa @t t *)

apply (elim conjE exE)
apply (rule disjI2)
apply (rule_tac x="sa" in exI)
apply (rule_tac x="t" in exI)
apply (simp add: dom_F_def)
apply (rule CF1_propI)
apply (simp_all)
done

(*********************************************************
               Lemma (Sequencesf in Dom_SF)
 *********************************************************)

(*******************************
             T2
 *******************************)

(*** lm 1 ***)

lemma Sequencesf_in_T2_lm1:
   "[| TF1 : dom_SF; TF2 : dom_SF; notick s;
          EX sa. s @t [Tick]t = rmtick sa & sa :t fst TF1 |]
       ==> (s, Evset)
           : (Rep_dom_T (fst TF1),Rep_dom_F (snd TF1)) ;;f Rep_dom_F (snd TF2) &
           (s @t [Tick]t, EvsetTick)
           : (Rep_dom_T (fst TF1),Rep_dom_F (snd TF1)) ;;f Rep_dom_F (snd TF2)"

apply (elim conjE exE)
apply (insert rmtick_is_notick_list)
apply (drule_tac x="Rep_trace sa" in spec)

apply (insert Abs_Rep_t)
apply (drule_tac x="rmtick_list (Rep_trace sa)" in spec)
apply (drule_tac x="s @t [Tick]t" in spec)
apply (simp add: rmtick_list_in_trace)

apply (rotate_tac 3)
apply (drule sym)
apply (simp)

apply (insert trace_app_in_trace_if)
apply (drule_tac x="s" in spec)
apply (drule_tac x="[Tick]t" in spec)
apply (simp)
apply (simp add: Abs_trace_inverse event_in_trace)
done

(*** main ***)

lemma Sequencesf_in_T2 :
  "ALL TF1 TF2. TF1:dom_SF & TF2:dom_SF --> CT2 (TF1 ;;sf TF2)"
apply (intro allI impI)
apply (simp add: Sequencesf_def CT2_def)
apply (intro allI impI)
apply (simp add: Sequencet_in Rep_dom_T Abs_dom_T_inverse)
apply (simp add: Rep_dom_T Rep_dom_F Sequencef_in Abs_dom_F_inverse)

apply (simp add: Sequencet_def)
apply (elim conjE)
apply (erule disjE)

  (* s @t [Tick]t = rmtick sa --> contra *)

apply (simp add: Sequencesf_in_T2_lm1)

apply (elim conjE exE)
apply (rename_tac TF1 TF2 s s1 s2)

apply (insert trace_last_nil_or_unnil)
apply (drule_tac x="s2" in spec)
apply (erule disjE)

 (* s2 = []t --> contra *)

apply (simp add: Rep_trace_inverse)
apply (simp add: appt_to_app_or_sym)
apply (simp add: rep_appt_to_apprep)
apply (simp add: nil_in_trace event_in_trace Abs_trace_inverse)

apply (rotate_tac 3)
apply (drule sym)
apply (simp)

 (* s2 = sa @t [a]t *)

apply (elim exE)
apply (simp add: app_trace_notick_ass)
apply (insert app_trace_same_last)
apply (drule_tac x="s" in spec)
apply (drule_tac x="Tick" in spec)
apply (drule_tac x="s1 @t sa" in spec)
apply (drule_tac x="a" in spec)
apply (simp add: app_trace_notick_notick)

apply (simp add: dom_SF_def CT2_def)

apply (elim conjE)
apply (rotate_tac 4)
apply (drule_tac x="sa" in spec)
apply (simp)

   (* Evset *) 

apply (rule conjI)
apply (simp add: Sequencef_def)
apply (rule disjI2)
apply (rule_tac x="s1" in exI)
apply (rule_tac x="sa" in exI)
apply (simp)

   (* EvsetTick *) 

apply (simp add: Sequencef_def)
apply (rule disjI2)
apply (rule_tac x="s1" in exI)
apply (rule_tac x="sa @t [Tick]t" in exI)
apply (simp add: app_trace_notick_ass)
done

(*******************************
             F2
 *******************************)

(*** lm 1 ***)

lemma Sequencesf_in_F2_lm1 :
  "[| TF1 : dom_SF; TF2 : dom_SF; notick s;
          ALL a. a : Y --> s @t [a]t ~: Rep_dom_T (fst TF1) ;;t Rep_dom_T (fst TF2);
          notick s & (s, insert Tick X) :f snd TF1 |]
       ==> (s, insert Tick (X Un Y)) :f snd TF1"
apply (elim conjE)

apply (simp add: dom_SF_def CF2_def)
apply (elim conjE)
apply (drule_tac x="s" in spec)
apply (drule_tac x="insert Tick X" in spec)
apply (drule_tac x="Y-{Tick}" in spec)
apply (simp)

apply (drule mp)
apply (intro allI impI)
apply (drule_tac x="a" in spec)
apply (simp add: Sequencet_def)
apply (elim conjE)

apply (rotate_tac -2)
apply (drule_tac x="s @t [a]t" in spec)

apply (insert rmtick_notick)
apply (rotate_tac -1)
apply (drule_tac x="s @t [a]t" in spec)
apply (rotate_tac -1)
apply (drule mp)
apply (simp add: trace_app_in_trace_if Abs_trace_inverse)
apply (simp add: event_in_trace Abs_trace_inverse)
apply (force)

apply (simp)

apply (subgoal_tac "insert Tick (X Un (Y - {Tick})) = insert Tick (X Un Y)")
apply (auto)
done

(*** main ***)

lemma Sequencesf_in_F2 :
  "ALL TF1 TF2. TF1:dom_SF & TF2:dom_SF --> CF2 (TF1 ;;sf TF2)"
apply (intro allI impI)
apply (simp add: Sequencesf_def CF2_def)
apply (intro allI impI)
apply (simp add: Sequencet_in Rep_dom_T Abs_dom_T_inverse)
apply (simp add: Rep_dom_T Rep_dom_F Sequencef_in Abs_dom_F_inverse)

apply (simp add: Sequencef_def)
apply (elim conjE)

apply (erule disjE)
apply (simp add: Sequencesf_in_F2_lm1)

apply (elim conjE exE)
apply (rename_tac TF1 TF2 s X Y s1 s2)

apply (rule disjI2)
apply (rule_tac x="s1" in exI)
apply (rule_tac x="s2" in exI)

apply (simp add: dom_SF_def CF2_def)
apply (elim conjE)
apply (rotate_tac -2)
apply (drule_tac x="s2" in spec)
apply (drule_tac x="X" in spec)
apply (drule_tac x="Y" in spec)
apply (simp)
apply (drule mp)

apply (insert app_trace_notick_notick_inv)
apply (rotate_tac -1)
apply (drule_tac x="s1" in spec)
apply (rotate_tac -1)
apply (drule_tac x="s2" in spec)
apply (simp)

apply (intro allI impI)
apply (drule_tac x="a" in spec)
apply (simp add: Sequencet_def)
apply (elim conjE)

apply (rotate_tac -1)
apply (drule_tac x="s1" in spec)
apply (simp)

apply (rotate_tac -1)
apply (drule_tac x="s2 @t [a]t" in spec)
apply (simp add: app_trace_notick_ass)

apply (simp)
done

(*******************************
             F3
 *******************************)

lemma Sequencesf_in_F3 :
  "ALL TF1 TF2. TF1:dom_SF & TF2:dom_SF --> CF3 (TF1 ;;sf TF2)"
apply (intro allI impI)
apply (simp add: Sequencesf_def CF3_def)
apply (intro allI impI)
apply (simp add: Sequencet_in Rep_dom_T Abs_dom_T_inverse)
apply (simp add: Rep_dom_T Rep_dom_F Sequencef_in Abs_dom_F_inverse)

apply (simp add: Sequencef_def)
apply (elim conjE exE)

apply (erule disjE)

 (* (s, insert Tick X) :f snd TF1 *)
apply (simp add: Sequencet_def)
apply (rule disjI1)
apply (rule_tac x="s" in exI)
apply (simp add: rmtick_notick)

apply (simp add: dom_SF_def CF3_def)
apply (erule conjE)+
apply (drule_tac x="s" in spec)
apply (drule mp)

apply (rule_tac x="insert Tick X" in exI)
apply (simp)
apply (simp)

 (* s = sa @t t *)
apply (elim conjE exE)
apply (simp add: Sequencet_def)
apply (rule disjI2)
apply (rule_tac x="sa" in exI)
apply (rule_tac x="t" in exI)
apply (simp)

apply (simp add: dom_SF_def CF3_def)
apply (erule conjE)+
apply (rotate_tac -1)
apply (drule_tac x="t" in spec)
apply (drule mp)
apply (rule_tac x="X" in exI)
apply (simp)
apply (simp)
done

(*******************************
      Sequencesf in dom_SF
 *******************************)

lemma Sequencesf_in :
  "ALL X TF1 TF2. TF1:dom_SF & TF2:dom_SF --> (TF1 ;;sf TF2): dom_SF"
apply (intro allI impI)
apply (simp (no_asm) add: dom_SF_def)
apply (simp add: Sequencesf_in_T2)
apply (simp add: Sequencesf_in_F2)
apply (simp add: Sequencesf_in_F3)
done

(*********************************************************
                     Lemmas (mono)
 *********************************************************)

(*** T ***)

lemma Sequencet_mono: 
  "ALL S1 T1 S2 T2. S1 <= T1 & S2 <= T2 --> S1 ;;t S2 <= T1 ;;t T2"
apply (intro allI impI)
apply (simp add: Sequencet_def)
apply (simp add: subset_def)
apply (intro allI impI ballI)
apply (elim conjE exE)
apply (simp)

 (* x = rmtick s *)
apply (erule disjE)
apply (elim conjE exE)
apply (drule_tac x="s" in bspec)
apply (simp)
apply (rule disjI1)
apply (rule_tac x="s" in exI)
apply (simp)

 (* x = s @t t *)
apply (elim conjE exE)
apply (drule_tac x="s @t [Tick]t" in bspec)
apply (simp)
apply (drule_tac x="t" in bspec)
apply (simp)
apply (rule disjI2)
apply (rule_tac x="s" in exI)
apply (rule_tac x="t" in exI)
apply (simp)
done

(*** F ***)

lemma Sequencef_mono: 
  "ALL S1 E1 T1 F1 E2 F2. S1 <= T1 & E1 <= F1 & E2 <= F2 
      --> ((S1, E1) ;;f E2 <= (T1, F1) ;;f F2)"

apply (intro allI impI)
apply (simp add: Sequencef_def del: split_eta_SetCompr2)
apply (rule subsetI)
apply (simp del: split_eta_SetCompr2)

 (* x = rmtick s *)
apply (erule disjE)
apply (elim conjE exE)
apply (rule disjI1)
apply (simp add: subset_F_def)
apply (force)

 (* x = (s @t t, X) *)
apply (elim conjE exE)
apply (rule disjI2)
apply (rule_tac x="s" in exI)
apply (rule_tac x="t" in exI)
apply (rule_tac x="X" in exI)

apply (simp add: subset_T_def)
apply (force)
done

(*** SF ***)

lemma Sequencesf_mono: 
  "ALL SE1 TF1 SE2 TF2. SE1 <= TF1 & SE2 <= TF2 
      --> SE1 ;;sf SE2 <= TF1 ;;sf TF2"

apply (simp add: Sequencesf_def del: split_paired_All)
apply (intro allI impI)
apply (simp (no_asm) add: order_pair_def)
apply (rule conjI)

 (* T *)
apply (simp add: subset_T_def order_pair_def)
apply (simp add: Sequencet_in Rep_dom_T Abs_dom_T_inverse)
apply (simp add: Sequencet_mono)

 (* F *)
apply (simp add: subset_T_def subset_F_def order_pair_def)
apply (simp add: Sequencef_in Rep_dom_T Rep_dom_F Abs_dom_F_inverse)
apply (erule conjE)
apply (simp add: Sequencef_mono)
done

(*********************************************************
                   Sequence trace eval
 *********************************************************)

lemma Sequencet_eval:
  "ALL P Q ev. [[P ;; Q]]T ev = [[P]]T ev ;;t [[Q]]T ev"
apply (intro allI)
apply (simp add: eval_T_def)
apply (simp add: eval_TF_decompo)
apply (simp add: Sequencesf_def)
apply (simp add: Sequencet_in Rep_dom_T Abs_dom_T_inverse)
done

(*********************************************************
                   Sequence failure eval
 *********************************************************)

lemma Sequencef_eval:
  "ALL X P Q ev. [[P ;; Q]]F ev = ([[P]]T ev, [[P]]F ev) ;;f [[Q]]F ev"
apply (intro allI)
apply (simp add: eval_T_def eval_F_def)
apply (simp add: eval_TF_decompo)
apply (simp add: Sequencesf_def)
apply (simp add: Sequencef_in Sequencet_in Rep_dom_T Rep_dom_F
                 Abs_dom_F_inverse Abs_dom_T_inverse)
done

end
