           (*-------------------------------------------*
            |                Csp-Prover                 |
            |                August 2004                |
            |        Yoshinao Isobe (AIST JAPAN)        |
            *-------------------------------------------*)

theory Csp_law_decompo = Csp_law_basic:

(*****************************************************************

         1. congruence, monotonicity

 *****************************************************************)

(*------------------------------------------------*
 |                                                |
 |              laws for monotonicity             |
 |                (Csp-Prover rule)               |
 |                                                |
 *------------------------------------------------*)

(*********************************************************
                        Prefix mono
 *********************************************************)

lemma csp_Prefix_mono:
  "[| a = b ; LET df IN P <=F[af] LET df IN Q |]
     ==> LET df IN a -> P <=F[af] LET df IN b -> Q"
apply (simp add: refsf_iff ref_Prefix_mono)
done

(*********************************************************
                    Prefix choice mono 
 *********************************************************)

lemma csp_Pre_choice_mono:
  "[| X = Y ; !! a. a:Y ==> LET df IN Pf a <=F[af] LET df IN Qf a |]
      ==> LET df IN ? :X -> Pf <=F[af] LET df IN ? :Y -> Qf"
apply (simp add: refsf_iff ref_Pre_choice_mono)
done

(*********************************************************
                      Ext choice mono
 *********************************************************)

lemma csp_Ext_choice_mono:
  "[| LET df IN P1 <=F[af] LET df IN Q1 ; 
      LET df IN P2 <=F[af] LET df IN Q2 |]
           ==> LET df IN P1 [+] P2 <=F[af] LET df IN Q1 [+] Q2"
apply (simp add: refsf_iff ref_Ext_choice_mono)
done

(*********************************************************
                      Int choice mono
 *********************************************************)

lemma csp_Int_choice_mono:
  "[| LET df IN P1 <=F[af] LET df IN Q1 ; 
      LET df IN P2 <=F[af] LET df IN Q2 |]
           ==> LET df IN P1 |~| P2 <=F[af] LET df IN Q1 |~| Q2"
apply (simp add: refsf_iff ref_Int_choice_mono)
done

(*********************************************************
                    Internal bind mono 
 *********************************************************)

lemma csp_Int_bind_mono:
  "[| X = Y ; !! a. a:Y ==> LET df IN Pf a <=F[af] LET df IN Qf a |]
      ==> LET df IN ! :X .. Pf <=F[af] LET df IN ! :X .. Qf"
apply (simp add: refsf_iff ref_Int_bind_mono)
done

(*********************************************************
                     Condition mono
 *********************************************************)

lemma csp_IF_mono:
  "[| b1 = b2 ;
      LET df IN P1 <=F[af] LET df IN Q1 ; 
      LET df IN P2 <=F[af] LET df IN Q2 |]
           ==> LET df IN IF b1 THEN P1 ELSE P2 <=F[af] 
               LET df IN IF b2 THEN Q1 ELSE Q2"
apply (simp add: refsf_iff ref_IF_mono)
done

(*********************************************************
                     Parallel mono 
 *********************************************************)

lemma csp_Parallel_mono:
  "[| X = Y ;
      LET df IN P1 <=F[af] LET df IN Q1 ; 
      LET df IN P2 <=F[af] LET df IN Q2 |]
           ==> LET df IN P1 |[X]| P2 <=F[af] LET df IN Q1 |[Y]| Q2"
apply (simp add: refsf_iff ref_Parallel_mono)
done

(*********************************************************
                        Hide mono
 *********************************************************)

lemma csp_Hide_mono:
  "[| X = Y ; LET df IN P <=F[af] LET df IN Q |]
    ==> LET df IN P -- X <=F[af] LET df IN Q -- Y"
apply (simp add: refsf_iff ref_Hide_mono)
done

(*********************************************************
                        Rename mono
 *********************************************************)

lemma csp_Rename_mono:
  "[| r1 = r2 ; LET df IN P <=F[af] LET df IN Q |]
   ==> LET df IN P [[r1]] <=F[af] LET df IN Q [[r2]]"
apply (simp add: refsf_iff ref_Rename_mono)
done

(*********************************************************
                       Sequence mono 
 *********************************************************)

lemma csp_Sequence_mono:
  "[| LET df IN P1 <=F[af] LET df IN Q1 ; 
      LET df IN P2 <=F[af] LET df IN Q2 |]
           ==> LET df IN P1 ;; P2 <=F[af] LET df IN Q1 ;; Q2"
apply (simp add: refsf_iff ref_Sequence_mono)
done

(*********************************************************
                        Timeout mono
 *********************************************************)

lemma csp_Timeout_mono:
  "[| LET df IN P1 <=F[af] LET df IN Q1 ;
      LET df IN P2 <=F[af] LET df IN Q2 |]
      ==> LET df IN P1 [> P2 <=F[af] LET df IN Q1 [> Q2"
apply (rule csp_Ext_choice_mono)
apply (rule csp_Int_choice_mono)
apply (simp_all add: csp_reflex)
done

(*------------------------------------------------*
 |                                                |
 |           laws for monotonicity (all)          |
 |                (Csp-Prover rule)               |
 |                                                |
 *------------------------------------------------*)

lemmas csp_free_mono =
   csp_Ext_choice_mono csp_Int_choice_mono csp_Parallel_mono
   csp_Hide_mono csp_Rename_mono csp_Sequence_mono

lemmas csp_mono = csp_free_mono
   csp_Prefix_mono csp_Pre_choice_mono 
   csp_Int_bind_mono csp_IF_mono 

lemmas csp_free_mono_left  = csp_free_mono[THEN csp_rule_tr_left_ref]
lemmas csp_free_mono_right = csp_free_mono[THEN csp_rule_tr_right_ref]

lemmas csp_mono_left  = csp_mono[THEN csp_rule_tr_left_ref]
lemmas csp_mono_right = csp_mono[THEN csp_rule_tr_right_ref]

lemmas csp_free_mono_af_left  = csp_free_mono[THEN csp_rule_af_left_ref]
lemmas csp_free_mono_af_right = csp_free_mono[THEN csp_rule_af_right_ref]

lemmas csp_mono_af_left  = csp_mono[THEN csp_rule_af_left_ref]
lemmas csp_mono_af_right = csp_mono[THEN csp_rule_af_right_ref]

(*------------------------------------------------*
 |                                                |
 |              laws for congruence               |
 |                (Csp-Prover rule)               |
 |                                                |
 *------------------------------------------------*)

(*********************************************************
                        Prefix cong
 *********************************************************)

lemma csp_Prefix_cong:
  "[| a = b ; LET df IN P =F[af] LET df IN Q |]
     ==> LET df IN a -> P =F[af] LET df IN b -> Q"
apply (simp add: eqsf_iff eq_Prefix_cong)
done

(*********************************************************
                    Prefix choice cong 
 *********************************************************)

lemma csp_Pre_choice_cong:
  "[| X = Y ; !! a. a:Y ==> LET df IN Pf a =F[af] LET df IN Qf a |]
      ==> LET df IN ? :X -> Pf =F[af] LET df IN ? :Y -> Qf"
apply (simp add: eqsf_iff eq_Pre_choice_cong)
done

(*********************************************************
                      Ext choice cong
 *********************************************************)

lemma csp_Ext_choice_cong:
  "[| LET df IN P1 =F[af] LET df IN Q1 ; 
      LET df IN P2 =F[af] LET df IN Q2 |]
           ==> LET df IN P1 [+] P2 =F[af] LET df IN Q1 [+] Q2"
apply (simp add: eqsf_iff eq_Ext_choice_cong)
done

(*********************************************************
                      Int choice cong
 *********************************************************)

lemma csp_Int_choice_cong:
  "[| LET df IN P1 =F[af] LET df IN Q1 ; 
      LET df IN P2 =F[af] LET df IN Q2 |]
           ==> LET df IN P1 |~| P2 =F[af] LET df IN Q1 |~| Q2"
apply (simp add: eqsf_iff eq_Int_choice_cong)
done

(*********************************************************
                    Internal bind cong 
 *********************************************************)

lemma csp_Int_bind_cong:
  "[| X = Y ; !! a. a:Y ==> LET df IN Pf a =F[af] LET df IN Qf a |]
      ==> LET df IN ! :X .. Pf =F[af] LET df IN ! :X .. Qf"
apply (simp add: eqsf_iff eq_Int_bind_cong)
done

(*********************************************************
                     Condition cong
 *********************************************************)

lemma csp_IF_cong:
  "[| b1 = b2 ;
      LET df IN P1 =F[af] LET df IN Q1 ; 
      LET df IN P2 =F[af] LET df IN Q2 |]
           ==> LET df IN IF b1 THEN P1 ELSE P2 =F[af] 
               LET df IN IF b2 THEN Q1 ELSE Q2"
apply (simp add: eqsf_iff eq_IF_cong)
done

(*********************************************************
                     Parallel cong 
 *********************************************************)

lemma csp_Parallel_cong:
  "[| X = Y ;
      LET df IN P1 =F[af] LET df IN Q1 ; 
      LET df IN P2 =F[af] LET df IN Q2 |]
           ==> LET df IN P1 |[X]| P2 =F[af] LET df IN Q1 |[Y]| Q2"
apply (simp add: eqsf_iff eq_Parallel_cong)
done

(*********************************************************
                        Hide cong
 *********************************************************)

lemma csp_Hide_cong:
  "[| X = Y ; LET df IN P =F[af] LET df IN Q |]
    ==> LET df IN P -- X =F[af] LET df IN Q -- Y"
apply (simp add: eqsf_iff eq_Hide_cong)
done

(*********************************************************
                        Rename cong
 *********************************************************)

lemma csp_Rename_cong:
  "[| r1 = r2 ; LET df IN P =F[af] LET df IN Q |]
   ==> LET df IN P [[r1]] =F[af] LET df IN Q [[r2]]"
apply (simp add: eqsf_iff eq_Rename_cong)
done

(*********************************************************
                       Sequence cong 
 *********************************************************)

lemma csp_Sequence_cong:
  "[| LET df IN P1 =F[af] LET df IN Q1 ; 
      LET df IN P2 =F[af] LET df IN Q2 |]
           ==> LET df IN P1 ;; P2 =F[af] LET df IN Q1 ;; Q2"
apply (simp add: eqsf_iff eq_Sequence_cong)
done

(*********************************************************
                        Timeout cong
 *********************************************************)

lemma csp_Timeout_cong:
  "[| LET df IN P1 =F[af] LET df IN Q1 ;
      LET df IN P2 =F[af] LET df IN Q2 |]
      ==> LET df IN P1 [> P2 =F[af] LET df IN Q1 [> Q2"
apply (rule csp_Ext_choice_cong)
apply (rule csp_Int_choice_cong)
apply (simp_all add: csp_reflex)
done

(*------------------------------------------------*
 |                                                |
 |            laws for congruence (all)           |
 |                (Csp-Prover rule)               |
 |                                                |
 *------------------------------------------------*)

lemmas csp_free_cong =
   csp_Ext_choice_cong csp_Int_choice_cong csp_Parallel_cong
   csp_Hide_cong csp_Rename_cong csp_Sequence_cong

lemmas csp_cong = csp_free_cong
   csp_Prefix_cong csp_Pre_choice_cong 
   csp_Int_bind_cong csp_IF_cong 

lemmas csp_free_cong_left  = csp_free_cong[THEN csp_rule_tr_left_eq]
lemmas csp_free_cong_right = csp_free_cong[THEN csp_rule_tr_right_eq]

lemmas csp_cong_left  = csp_cong[THEN csp_rule_tr_left_eq]
lemmas csp_cong_right = csp_cong[THEN csp_rule_tr_right_eq]

lemmas csp_free_cong_af_left  = csp_free_cong[THEN csp_rule_af_left_eq]
lemmas csp_free_cong_af_right = csp_free_cong[THEN csp_rule_af_right_eq]

lemmas csp_cong_af_left  = csp_cong[THEN csp_rule_af_left_eq]
lemmas csp_cong_af_right = csp_cong[THEN csp_rule_af_right_eq]

(*------------------------------------------------*
 |                                                |
 |           laws for decompo (ALL)               |
 |                (Csp-Prover rule)               |
 |                                                |
 *------------------------------------------------*)

lemmas csp_free_decompo = csp_free_mono csp_free_cong
lemmas csp_decompo = csp_mono csp_cong

lemmas csp_free_decompo_left  = csp_free_mono_left csp_free_cong_left
lemmas csp_free_decompo_right = csp_free_mono_right csp_free_cong_right

lemmas csp_decompo_left  = csp_mono_left csp_cong_left
lemmas csp_decompo_right = csp_mono_right csp_cong_right


lemmas csp_free_decompo_af_left  = csp_free_mono_af_left csp_free_cong_af_left
lemmas csp_free_decompo_af_right = csp_free_mono_af_right csp_free_cong_af_right

lemmas csp_decompo_af_left  = csp_mono_af_left csp_cong_af_left
lemmas csp_decompo_af_right = csp_mono_af_right csp_cong_af_right

(*------------------------------------------------*
 |                                                |
 |              remove heads of hnf               |
 |                (Csp-Prover rule)               |
 |                                                |
 *------------------------------------------------*)

lemmas csp_rm_head_mono = csp_Prefix_mono csp_Pre_choice_mono 
lemmas csp_rm_head_cong = csp_Prefix_cong csp_Pre_choice_cong
lemmas csp_rm_head      = csp_rm_head_mono csp_rm_head_cong

end
