           (*-------------------------------------------*
            |                Csp-Prover                 |
            |                July 2004                  |
            |        Yoshinao Isobe (AIST JAPAN)        |
            *-------------------------------------------*)

theory Csp_law_fp = Csp_law_basic:

(*****************************************************************

         1. name unwinding
         2. fixed point induction

 *****************************************************************)

(*********************************************************
                 name unwinding function
 *********************************************************)

(* [| !! C. nohide (df C) ; !! C. guard  (df C) |] *)
(* ==> LET df IN <C> =F[af] LET df IN df C             *)

lemmas csp_unwind = eq_unwind

lemmas csp_unwind_rw_left_eq   = csp_unwind[THEN csp_rule_rw_left_eq]
lemmas csp_unwind_rw_left_ref  = csp_unwind[THEN csp_rule_rw_left_ref]
lemmas csp_unwind_rw_right_eq  = csp_unwind[THEN csp_rule_rw_right_eq]
lemmas csp_unwind_rw_right_ref = csp_unwind[THEN csp_rule_rw_right_ref]

lemmas csp_unwind_rw_left      = csp_unwind_rw_left_eq csp_unwind_rw_left_ref
lemmas csp_unwind_rw_right     = csp_unwind_rw_right_eq csp_unwind_rw_right_ref

lemmas csp_unwind_rw           = csp_unwind_rw_left csp_unwind_rw_right

(*-------------------------------------------------------*
 |                                                       |
 |        Check fixpoint  (Csp-Prover simp rule)         |
 |                                                       |
 *-------------------------------------------------------*)

(*** unfold fixpoint for equality ***)

lemma csp_check_eq:
  "[| !!C. nohide ((LetD R1) C) ; !!C. guard ((LetD R1) C) ;
      !!C. nohide ((LetD R2) C) ; !!C. guard ((LetD R2) C) ;
      (LET (LetD R2) IN (Rewrite (InP R1) By f12) =F 
       LET (LetD R2) IN (InP R2)) ;
      !!C. LET (LetD R2) IN (Rewrite (LetD R1 C) By f12) =F
           LET (LetD R2) IN (f12 C) |]
    ==> CHECK R1 =F R2 BY f12"
apply (simp add: CHECKeq_def)
done

(*** unfold fixpoint for refinement ***)

lemma csp_check_ref:
  "[| !!C. nohide ((LetD R1) C) ; !!C. guard ((LetD R1) C) ;
      !!C. nohide ((LetD R2) C) ; !!C. guard ((LetD R2) C) ;
      (LET (LetD R2) IN (Rewrite (InP R1) By f12) <=F 
       LET (LetD R2) IN (InP R2)) ;
      !!C. LET (LetD R2) IN (Rewrite (LetD R1 C) By f12) <=F
           LET (LetD R2) IN (f12 C) |]
    ==> CHECK R1 <=F R2 BY f12"
apply (simp add: CHECKref_def)
done

(*** unfold fixpoint for reverse refinement ***)

lemma csp_check_rev:
  "[| !!C. nohide ((LetD R1) C) ; !!C. guard ((LetD R1) C) ;
      !!C. nohide ((LetD R2) C) ; !!C. guard ((LetD R2) C) ;
     (LET (LetD R2) IN (InP R2) <=F 
      LET (LetD R2) IN (Rewrite (InP R1) By f12)) ;
      !!C. LET (LetD R2) IN (f12 C) <=F 
           LET (LetD R2) IN (Rewrite (LetD R1 C) By f12) |]
    ==> CHECK R1 =>F R2 BY f12"
apply (simp add: CHECKrev_def)
done

lemmas csp_check = csp_check_eq csp_check_ref csp_check_rev

(*-------------------------------------------------------*
 |                                                       |
 |    Fixed point induction  (Csp-Prover intro rule)     |
 |                                                       |
 *-------------------------------------------------------*)

lemma csp_fp_induct_eq:
  "CHECK R1 =F R2 BY f12 ==> R1 =F R2"
apply (insert eq_fp_induct[of "LetD R1" "InP R1" "LetD R2" "InP R2" f12])
apply (simp add: LetD_InP)
done

lemma csp_fp_induct_ref:
  "CHECK R1 <=F R2 BY f12 ==> R1 <=F R2"
apply (insert ref_fp_induct[of "LetD R1" "InP R1" "LetD R2" "InP R2" f12])
apply (simp add: LetD_InP)
done

lemma csp_fp_induct_rev:
  "CHECK R1 =>F R2 BY f12 ==> R2 <=F R1"
apply (insert rev_fp_induct[of "LetD R1" "InP R1" "LetD R2" "InP R2" f12])
apply (simp add: LetD_InP)
done

lemmas csp_fp_induct = csp_fp_induct_eq csp_fp_induct_ref

end
