           (*-------------------------------------------*
            |                Csp-Prover                 |
            |                July 2004                  |
            |        Yoshinao Isobe (AIST JAPAN)        |
            *-------------------------------------------*)

theory Csp_map_proc = Csp_proc + Csp_map_Prefix + Csp_map_Ext_choice + 
                      Csp_map_Int_choice + Csp_map_Parallel +
                      Csp_map_Rename + Csp_map_Sequence:

(*  The following simplification rules are deleted in this theory file *)
(*  because they unexpectly rewrite UnionT and InterT.                 *)
(*                  Union (B ` A) = (UN x:A. B x)                      *)
(*                  Inter (B ` A) = (INT x:A. B x)                     *)

declare Union_image_eq [simp del]
declare Inter_image_eq [simp del]

   (* note: split_paired_All: (ALL x. ?P x) = (ALL a b. ?P (a, b)) *)
   (*       split_paired_Ex : (EX x. ?P x) = (EX a b. ?P (a, b))   *)

declare split_paired_All [simp del]
declare split_paired_Ex  [simp del]

(*****************************************************************

         1. proc : nohide --> non-expanding
         2. proc : guarded & nohide --> contraction
         3. 
         4. 

 *****************************************************************)

(*********************************************************
                      Prefix map
 *********************************************************)

lemma contract_Prefix:
  "ALL a P. non_expanding [[P]]SF 
        --> contraction_alpha [[a -> P]]SF (1/2)"
apply (simp add: non_expanding_def contraction_alpha_def 
                 map_alpha_def eval_SF_def)
apply (simp add: distance_dom_SF Prefixsf_in procsf_in Abs_dom_SF_inverse)
apply (intro allI impI)
apply (drule_tac x="x" in spec)
apply (drule_tac x="y" in spec)
apply (simp add: dis_Prefixsf)
done

lemma non_expand_Prefix:
  "ALL a P. non_expanding [[P]]SF --> non_expanding [[a -> P]]SF"
apply (intro allI impI)
apply (insert contract_Prefix)
apply (drule_tac x="a" in spec)
apply (drule_tac x="P" in spec)
apply (simp add: contraction_alpha_contraction
                 contraction_non_expanding)
done

(*********************************************************
                   Prefix choice map
 *********************************************************)

lemma contract_Pre_choice:
  "ALL X Pf. (ALL a. non_expanding [[Pf a]]SF)
        --> contraction_alpha [[? :X -> Pf]]SF (1/2)"
apply (simp add: non_expanding_def contraction_alpha_def 
                 map_alpha_def eval_SF_def)
apply (simp add: distance_dom_SF)
apply (simp add: Pre_choicesf_in_induct procsf_in Abs_dom_SF_inverse)
apply (intro allI impI)

apply (insert dis_Ext_choicesf)
apply (drule_tac x=
       "insert (STOPsf,STOPsf) 
        {(a ->sf [[Pf a]]TF x, a ->sf [[Pf a]]TF y) |a. a : X}" in spec)
apply (simp add: image_Collect_fun)
apply (elim conjE exE)
apply (erule disjE)

      (* sup by STOP *)

apply (simp add: same_pnt_zero positive_ms)
apply (insert positive_ms)
apply (drule_tac x="x" in spec)
apply (drule_tac x="y" in spec)
apply (simp)

      (* sup by Pf(a) *)

apply (simp add: positive_ms)  (* dummy *)
apply (elim conjE exE)
apply (rename_tac X Pf ev ev' SETF a)

apply (drule_tac x="a" in spec)
apply (drule_tac x="ev" in spec)
apply (drule_tac x="ev'" in spec)
apply (simp add: dis_Prefixsf)
done

lemma non_expand_Pre_choice:
  "ALL X Pf. (ALL a. non_expanding [[Pf a]]SF)
        --> non_expanding [[? :X -> Pf]]SF"
apply (intro allI impI)
apply (insert contract_Pre_choice)
apply (drule_tac x="X" in spec)
apply (drule_tac x="Pf" in spec)
apply (simp add: contraction_alpha_contraction
                 contraction_non_expanding)
done

(*********************************************************
                    Ext_choice map
 *********************************************************)

lemma map_Ext_choice:
  "ALL alpha P Q. map_alpha [[P]]SF alpha & map_alpha [[Q]]SF alpha
        --> map_alpha [[P [+] Q]]SF alpha"
apply (simp add: map_alpha_def eval_SF_def)
apply (simp add: distance_dom_SF Ext_choicesf_in procsf_in Abs_dom_SF_inverse)
apply (intro allI impI)
apply (erule conjE)+
apply (drule_tac x="x" in spec)
apply (drule_tac x="x" in spec)
apply (drule_tac x="y" in spec)
apply (drule_tac x="y" in spec)
apply (insert dis_Ext_choicesf)
apply (drule_tac x=
       "{([[P]]TF x, [[P]]TF y), ([[Q]]TF x, [[Q]]TF y)}" in spec)
apply (simp add: image_Collect_fun)
apply (auto)
done

(*********************************************************
                    Int_choicee map
 *********************************************************)

lemma map_Int_choice:
  "ALL alpha P Q. map_alpha [[P]]SF alpha & map_alpha [[Q]]SF alpha
        --> map_alpha [[P |~| Q]]SF alpha"
apply (simp add: map_alpha_def eval_SF_def)
apply (simp add: distance_dom_SF Int_choicesf_in procsf_in Abs_dom_SF_inverse)
apply (intro allI impI)
apply (erule conjE)+
apply (drule_tac x="x" in spec)
apply (drule_tac x="x" in spec)
apply (drule_tac x="y" in spec)
apply (drule_tac x="y" in spec)
apply (insert dis_Int_choicesf)
apply (drule_tac x=
       "{([[P]]TF x, [[P]]TF y), ([[Q]]TF x, [[Q]]TF y)}" in spec)
apply (simp add: image_Collect_fun)
apply (auto)
done

(*********************************************************
                  Internal Bind
 *********************************************************)

lemma map_Int_bind:
  "ALL alpha X Pf. (ALL a. map_alpha [[Pf a]]SF alpha)
        --> map_alpha [[! :X .. Pf]]SF alpha"
apply (simp add: map_alpha_def eval_SF_def)
apply (simp add: distance_dom_SF)
apply (simp add: Int_bindsf_in_induct
                 procsf_in Abs_dom_SF_inverse)
apply (intro allI impI)

apply (insert dis_Int_choicesf)
apply (drule_tac x=
       "insert (DIVsf,DIVsf) 
        {([[Pf a]]TF x, [[Pf a]]TF y) |a. a : X}" in spec)
apply (simp add: image_Collect_fun)
apply (elim conjE exE)
apply (erule disjE)

      (* sup by DIV *)

apply (simp add: same_pnt_zero positive_ms)
apply (insert positive_ms)
apply (drule_tac x="x" in spec)
apply (drule_tac x="y" in spec)
apply (insert ALL_real_mult_left_mono)
apply (drule_tac x="0" in spec)
apply (drule_tac x="distance (x, y)" in spec)
apply (drule_tac x="alpha" in spec)
apply (simp)

      (* sup by Pf(a) *)

apply (simp add: positive_ms)  (* dummy *)
apply (elim conjE exE)
apply (rename_tac alpha X Pf ev ev' SETF a)

apply (drule_tac x="a" in spec)
apply (drule_tac x="ev" in spec)
apply (drule_tac x="ev'" in spec)
apply (simp)
done

(*********************************************************
                      Parallel map
 *********************************************************)

lemma map_Parallel:
  "ALL alpha X P Q. map_alpha [[P]]SF alpha & map_alpha [[Q]]SF alpha
        --> map_alpha [[P |[X]| Q]]SF alpha"
apply (simp add: map_alpha_def eval_SF_def)
apply (simp add: distance_dom_SF Parallelsf_in procsf_in Abs_dom_SF_inverse)
apply (intro allI impI)
apply (erule conjE)+
apply (drule_tac x="x" in spec)
apply (drule_tac x="x" in spec)
apply (drule_tac x="y" in spec)
apply (drule_tac x="y" in spec)
apply (insert dis_Parallelsf)
apply (drule_tac x="X" in spec)
apply (drule_tac x="[[P]]TF x" in spec)
apply (drule_tac x="[[P]]TF y" in spec)
apply (drule_tac x="[[Q]]TF x" in spec)
apply (drule_tac x="[[Q]]TF y" in spec)

  (* solve max *)

apply (simp add: le_max_iff_disj)
apply (auto)
done

(*********************************************************
                      Rename map
 *********************************************************)

lemma map_Rename:
  "ALL alpha P r. map_alpha [[P]]SF alpha
        --> map_alpha [[P [[r]] ]]SF alpha"
apply (simp add: map_alpha_def eval_SF_def)
apply (simp add: distance_dom_SF Renamesf_in procsf_in Abs_dom_SF_inverse)
apply (intro allI impI)
apply (erule conjE)+
apply (drule_tac x="x" in spec)
apply (drule_tac x="y" in spec)
apply (insert dis_Renamesf)
apply (drule_tac x="r" in spec)
apply (drule_tac x="[[P]]TF x" in spec)
apply (drule_tac x="[[P]]TF y" in spec)
apply (simp)
done

(*********************************************************
                      Sequence map
 *********************************************************)

lemma map_Sequence:
  "ALL alpha P Q. map_alpha [[P]]SF alpha & map_alpha [[Q]]SF alpha
        --> map_alpha [[P ;; Q]]SF alpha"
apply (simp add: map_alpha_def eval_SF_def)
apply (simp add: distance_dom_SF Sequencesf_in procsf_in Abs_dom_SF_inverse)
apply (intro allI impI)
apply (erule conjE)+
apply (drule_tac x="x" in spec)
apply (drule_tac x="x" in spec)
apply (drule_tac x="y" in spec)
apply (drule_tac x="y" in spec)
apply (insert dis_Sequencesf)
apply (drule_tac x="[[P]]TF x" in spec)
apply (drule_tac x="[[P]]TF y" in spec)
apply (drule_tac x="[[Q]]TF x" in spec)
apply (drule_tac x="[[Q]]TF y" in spec)

  (* solve max *)

apply (simp add: le_max_iff_disj)
apply (auto)
done

(*********************************************************
                   Process Name map
 *********************************************************)

lemma non_expand_Proc_name:
  "ALL C. non_expanding [[<C>]]SF"
apply (simp add: non_expanding_def map_alpha_def eval_SF_def)
apply (simp add: Rep_dom_SF_inverse)
apply (intro allI)
apply (insert prod_distance)
apply (drule_tac x="x" in spec)
apply (drule_tac x="y" in spec)
apply (simp add: is_lub_def is_ub_def)
apply (erule conjE)
apply (drule_tac x="distance (x C, y C)" in spec)
apply (auto)
done

(*********************************************************
               !!!   proc non-expanding   !!!
 *********************************************************)

lemma non_expand_proc:
  "ALL P. nohide P --> non_expanding [[P]]SF"
apply (intro allI)
apply (induct_tac P)
apply (simp add: non_expanding_def map_alpha_def 
                 eval_SF_def same_pnt_zero positive_ms)
apply (simp add: non_expanding_def map_alpha_def 
                 eval_SF_def same_pnt_zero positive_ms)
apply (simp add: non_expanding_def map_alpha_def 
                 eval_SF_def same_pnt_zero positive_ms)
apply (simp add: non_expand_Prefix)
apply (simp add: non_expand_Pre_choice)
apply (simp add: non_expanding_def map_Ext_choice)
apply (simp add: non_expanding_def map_Int_choice)
apply (simp add: non_expanding_def map_Int_bind)
apply (simp add: non_expanding_def map_alpha_def eval_SF_def) 
                                                    (* if then else *)
apply (simp add: non_expanding_def map_Parallel)
apply (simp)                                        (* nohide *)
apply (simp add: non_expanding_def map_Rename)
apply (simp add: non_expanding_def map_Sequence)
apply (simp add: non_expand_Proc_name)
done

(*********************************************************
            !!!   proc contraction map   !!!
 *********************************************************)

lemma contract_proc:
  "ALL P. nohide P & guard P --> contraction_alpha [[P]]SF (1/(2::real))"
apply (intro allI)
apply (induct_tac P)
apply (simp add: contraction_alpha_def map_alpha_def 
                 eval_SF_def same_pnt_zero positive_ms)
apply (simp add: contraction_alpha_def map_alpha_def 
                 eval_SF_def same_pnt_zero positive_ms)
apply (simp add: contraction_alpha_def map_alpha_def 
                 eval_SF_def same_pnt_zero positive_ms)
apply (simp add: non_expand_proc contract_Prefix)
apply (simp add: non_expand_proc contract_Pre_choice)

apply (simp add: contraction_alpha_def map_Ext_choice)

apply (simp add: contraction_alpha_def map_Ext_choice)
apply (simp add: contraction_alpha_def map_Int_choice)
apply (simp add: contraction_alpha_def map_Int_bind)
apply (simp add: contraction_alpha_def map_alpha_def eval_SF_def)
                                                    (* if then else *)
apply (simp add: contraction_alpha_def map_Parallel)
apply (simp)                                        (* nohide *)
apply (simp add: contraction_alpha_def map_Rename)
apply (simp add: contraction_alpha_def map_Sequence)
apply (simp)                                        (* guard *)
done

(****************** to add them again ******************)

declare Union_image_eq [simp]
declare Inter_image_eq [simp]

declare split_paired_All [simp add]
declare split_paired_Ex  [simp add]

end
