           (*-------------------------------------------*
            |                Csp-Prover                 |
            |                June 2004                  |
            |        Yoshinao Isobe (AIST JAPAN)        |
            *-------------------------------------------*)

theory Domain_F = Trace + RS:

(*****************************************************************

         1. 
         2. 
         3. 
         4. 

 *****************************************************************)

(*  The following simplification rules are deleted in this theory file *)
(*  because they unexpectly rewrite UnionT and InterT.                 *)
(*                  Union (B ` A) = (UN x:A. B x)                      *)
(*                  Inter (B ` A) = (INT x:A. B x)                     *)

declare Union_image_eq [simp del]
declare Inter_image_eq [simp del]

(***********************************************************
                 type def (Failure Part)
 ***********************************************************)

types 'a failure = "'a trace * 'a event set"       (* synonym *)

consts
  CF1    :: "'a failure set => bool" 

defs
  CF1_def : "CF1 F == ALL s X Y. ((s,X) : F & Y <= X) --> (s,Y) : F"

typedef 'a dom_F = "{F::('a failure set). CF1(F)}"
apply (rule_tac x ="{([]t, {})}" in exI)
apply (simp add: CF1_def)
done

(*** CF1 property I (rule) ***)

lemma CF1_propI:
  "[| CF1 F; (s, X) : F; Y <= X|] ==> (s, Y) : F"
apply (unfold CF1_def)
apply (blast)
done

(*** Rep_dom_F s : dom_F ***)

lemma ALL_Rep_dom_F: "ALL F. Rep_dom_F F : dom_F"
apply (simp add: Rep_dom_F)
done

(***********************************************************
                   operators on dom_F
 ***********************************************************)

syntax
  "_mem_F"     :: "'a failure => 'a dom_F => bool"   ("(_/ :f _)" [50, 51] 50)
  "_nonmem_F"  :: "'a failure => 'a dom_F => bool"   ("(_/ ~:f _)" [50, 51] 50)
  "_Collect_F" :: "('a failure => bool) => 'a dom_F" ("CollectF")
  "_Union_F"   :: "'a dom_F set => 'a dom_F"         ("UnionF _" [90] 90)
  "_Un_F"      :: "'a dom_F => 'a dom_F => 'a dom_F" ("_ UnF _" [65,66] 65)
  "_Inter_F"   :: "'a dom_F set => 'a dom_F"         ("InterF _" [90] 90)
  "_Int_F"     :: "'a dom_F => 'a dom_F => 'a dom_F" ("_ IntF _" [70,71] 70)

  "@Coll_F"    :: "pttrn => bool => 'a dom_F"     ("(1{_./ _}f)")
  "@Finset_F"  :: "args => 'a dom_F"              ("{(_)}f")

  "_emp_F"     :: "'a dom_F"                      ("{}f")
  "_UNIV_F"    :: "'a dom_F"                      ("UNIVf")

translations
  "x :f F"     == "x : (Rep_dom_F F)"
  "x ~:f F"     == "x ~: (Rep_dom_F F)"
  "CollectF P" == "Abs_dom_F (Collect P)"
  "UnionF Fs"  == "Abs_dom_F (Union (Rep_dom_F ` Fs))"
  "F UnF E"    == "UnionF {F,E}"
  "InterF Fs"  == "Abs_dom_F (Inter (Rep_dom_F ` Fs))"
  "F IntF E"   == "InterF {F,E}"

  "{x. P}f"    == "Abs_dom_F {x. P}"
  "{X}f"       == "Abs_dom_F {X}"

  "{}f"        == "Abs_dom_F {}"
  "UNIVf"      == "Abs_dom_F UNIV"


(*********************************************************
          The relation (<=) is defined over dom_F
 *********************************************************)

instance dom_F :: (type) ord 
apply (intro_classes)
done

defs (overloaded)
  subset_F_def   :  "F <= E  ==  (Rep_dom_F F) <= (Rep_dom_F E)"
  psubset_F_def  :  "F <  E  ==  (Rep_dom_F F) <  (Rep_dom_F E)"

(*********************************************************
          The relation (<=) is a partial order
 *********************************************************)

instance dom_F :: (type) order
apply (intro_classes)
apply (unfold subset_F_def psubset_F_def)
apply (simp)
apply (erule order_trans, simp)
apply (drule order_antisym, simp)
apply (simp add: Rep_dom_F_inject)
apply (simp only: order_less_le Rep_dom_F_inject)
done

(***********************************************************
                          lemmas
 ***********************************************************)

(*******************************
             basic
 *******************************)


(*** F1 ***)

lemma dom_F_F1:
  "ALL s X Y F. F:dom_F & (s,X) : F & Y <= X --> (s,Y) : F"
apply (intro allI impI)
apply (simp add: dom_F_def)
apply (unfold CF1_def)
apply (erule conjE)+
apply (drule_tac x="s" in spec)
apply (drule_tac x="X" in spec)
apply (drule_tac x="Y" in spec)
apply (simp)
done

(*******************************
         check in dom_F
 *******************************)

(*** {}f ***)

lemma emptyset_in_dom_F: "{} : dom_F"
apply (simp add: dom_F_def CF1_def)
done

(*** [] (for STOP) ***)

lemma nil_t_in_dom_F: "{([]t, X) |X. X <= EvsetTick} : dom_F"
apply (auto simp add: dom_F_def CF1_def)
done

(*** [Tick] (for SKIP) ***)
lemma nil_Tick_t_in_dom_F: "{([]t, X) |X. X <= Evset} Un 
                            {([Tick]t, X) |X. X <= EvsetTick} : dom_F"
apply (auto simp add: dom_F_def CF1_def)
done

(*** Union ***)

lemma dom_F_Union_in_dom_F:
  "ALL Fs. (Union (Rep_dom_F ` Fs)) : dom_F"
apply (intro allI)
apply (simp add: dom_F_def CF1_def)
apply (intro allI impI)

apply (erule conjE)
apply (simp add: Bex_def)
apply (erule exE)
apply (rename_tac Fs s X Y F)
apply (rule_tac x="F" in exI)

apply (insert dom_F_F1)
apply (drule_tac x="s" in spec)
apply (drule_tac x="X" in spec)
apply (drule_tac x="Y" in spec)
apply (drule_tac x="Rep_dom_F F" in spec)
apply (simp add: Rep_dom_F)
done

(*** Un ***)

lemma dom_F_Un_in_dom_F:
  "ALL F E. (Rep_dom_F F Un Rep_dom_F E) : dom_F"
apply (intro allI)
apply (insert dom_F_Union_in_dom_F)
apply (drule_tac x="{F,E}" in spec)
apply (simp)
done

(*** Inter ***)

lemma dom_F_Inter_in_dom_F:
  "ALL Fs. (Inter (Rep_dom_F ` Fs)) : dom_F"
apply (intro allI)
apply (simp add: dom_F_def CF1_def)
apply (intro allI impI)

apply (erule conjE)
apply (simp add: Ball_def)
apply (intro allI impI)
apply (rename_tac Fs s X Y F)
apply (drule_tac x="F" in spec)

apply (insert dom_F_F1)
apply (drule_tac x="s" in spec)
apply (drule_tac x="X" in spec)
apply (drule_tac x="Y" in spec)
apply (drule_tac x="Rep_dom_F F" in spec)
apply (simp add: Rep_dom_F)
done

(*** Int ***)

lemma dom_F_Int_in_dom_F:
  "ALL F E. (Rep_dom_F F Int Rep_dom_F E) : dom_F"
apply (intro allI)
apply (insert dom_F_Inter_in_dom_F)
apply (drule_tac x="{F,E}" in spec)
apply (simp)
done

(*******************************
     dom_F type --> set type
 *******************************)

(*** UnionF ***)

lemma dom_F_UnionF_Rep:
  "ALL Fs. Rep_dom_F (UnionF Fs) = Union (Rep_dom_F ` Fs)"
apply (simp add: Abs_dom_F_inverse dom_F_Union_in_dom_F)
done

(*** UnF ***)

lemma dom_F_UnF_Rep:
  "ALL F E. Rep_dom_F (F UnF E) = (Rep_dom_F F) Un (Rep_dom_F E)"
apply (simp add: Abs_dom_F_inverse dom_F_Un_in_dom_F)
done

(*** InterF ***)

lemma dom_F_InterF_Rep:
  "ALL Fs. Rep_dom_F (InterF Fs) = Inter (Rep_dom_F ` Fs)"
apply (simp add: Abs_dom_F_inverse dom_F_Inter_in_dom_F)
done

(*** IntF ***)

lemma dom_F_IntF_Rep:
  "ALL F E. Rep_dom_F (F IntF E) = (Rep_dom_F F) Int (Rep_dom_F E)"
apply (simp add: Abs_dom_F_inverse dom_F_Int_in_dom_F)
done

(*********************************************************
                 Restriction in Dom_F
 *********************************************************)

instance dom_F :: (type) ms0
apply (intro_classes)
done

consts
  restFset :: "'a dom_F => nat => 'a failure set" ("_ restFset _" [55,56] 55)
  limitFset:: "'a dom_F infinite_seq => 'a failure set"

defs
 restFset_def  : 
   "F restFset n == {sX. EX s. (EX X. sX = (s, X)) & sX :f F &
               (lengtht s < n |
                lengtht s = n & (EX s'. s = s' @t [Tick]t & notick s'))}"

 limitFset_def : 
   "limitFset Fs == {sX. EX s. (EX X. sX = (s, X)) &
           (sX :f Fs (Suc (lengtht s)) |
            sX :f Fs (lengtht s) & (EX s'. s = s' @t [Tick]t & notick s'))}"

defs (overloaded)
  rest_F_def : "F rest n == Abs_dom_F (F restFset n)"

syntax
  "_limitF"  :: "'a dom_F infinite_seq => 'a dom_F"   ("limitF")

translations
  "limitF Fs" == "Abs_dom_F (limitFset Fs)"

(*********************************************************
              Lemmas (Restriction in Dom_F)
 *********************************************************)

(*** restFset_def in dom_F ***)

lemma restFset_def_in : "ALL F n. {r. EX s. (EX X. r = (s, X)) & r :f F &
          (lengtht s < n |
           lengtht s = n & (EX s'. s = s' @t [Tick]t & notick s'))}: dom_F"
apply (intro allI)
apply (simp add: dom_F_def CF1_def)
apply (intro allI impI)

apply (insert dom_F_F1)
apply (drule_tac x="s" in spec)
apply (drule_tac x="X" in spec)
apply (drule_tac x="Y" in spec)
apply (drule_tac x="Rep_dom_F F" in spec)
apply (simp add: Rep_dom_F)
done

(*** restFset in dom_F ***)

lemma restFset_in : "ALL F n. F restFset n : dom_F"
apply (simp add: restFset_def restFset_def_in)
done

(*** F rest = E rest --> F = E ***)

lemma rest_dom_F_eq: 
  "ALL (F::'a dom_F) (E::'a dom_F) sX.
      sX :f F & F restFset Suc (lengtht (fst sX)) = 
                E restFset Suc (lengtht (fst sX))
         --> sX :f E"
apply (auto simp add: restFset_def)
done


(*********************************************************
                     Dom_F --> RS
 *********************************************************)

(*******************************
        zero_eq_rs_dom_F
 *******************************)

(*** (contra) restFset 0 ***)

lemma contra_zero_dom_F: "ALL (F::'a dom_F) n. F restFset n ~= {} --> 0 < n"
apply (intro allI impI)
apply (simp add: restFset_def)
apply (erule exE)
apply (erule conjE)+
apply (erule disjE)
apply (simp)

apply (erule conjE)
apply (erule exE)+
apply (simp add: notick_app_trace_rep)
apply (simp add: tick_in_trace Abs_trace_inverse)
done

(*** restFset 0 ***)

lemma zero_dom_F: "ALL (F::'a dom_F). F restFset 0 = {}"
apply (intro allI)
apply (insert contra_zero_dom_F)
apply (drule_tac x="F" in spec)
apply (drule_tac x="0" in spec)
apply (simp)
done

(*** zero_eq_rs_dom_F ***)

lemma zero_eq_rs_dom_F: "ALL (F::'a dom_F) E. F rest 0 = E rest 0"
apply (simp add: rest_F_def)
apply (simp add: restFset_in Abs_dom_F_inject)
apply (simp add: zero_dom_F)
done

(*******************************
         min_rs_dom_F
 *******************************)

lemma min_rs_dom_F:
  "ALL (F::'a dom_F) (m::nat) (n::nat). 
     (F rest m) rest n = F rest (min m n)"
apply (intro allI)
apply (simp add: rest_F_def)
apply (simp add: restFset_in Abs_dom_F_inject)

apply (simp add: restFset_def)
apply (simp add: restFset_def_in Abs_dom_F_inverse)

apply (simp add: Collect_open)
apply (intro allI)
apply (rename_tac F m n s X)

apply (rule iffI)

apply (simp)
apply (erule conjE)+
apply (erule disjE)+
apply (simp)
apply (simp add: min_def)
apply (erule disjE)
apply (simp add: min_def)
apply (simp)

apply (simp)
apply (erule conjE)+
apply (erule disjE)
apply (simp)
apply (simp)

apply (rule conjI)

apply (case_tac "n < m")
apply (simp add: min_def)
apply (simp add: min_def)

apply (case_tac "m < n")
apply (simp_all add: min_def)
done

(*******************************
        diff_rs_dom_F
 *******************************)

(*** contra <= ***)

lemma contra_diff_rs_dom_F_left: 
  "ALL (F::'a dom_F) (E::'a dom_F).
    (ALL (n::nat). F rest n = E rest n) --> F <= E"
apply (intro allI impI)
apply (simp add: subset_F_def)

apply (rule subsetI)
apply (drule_tac x="Suc (lengtht (fst x))" in spec)
apply (simp add: rest_F_def) 
apply (simp add: restFset_in Abs_dom_F_inject)
apply (auto simp add: restFset_def)
done

(*** contra = ***)

lemma contra_diff_rs_dom_F: 
  "ALL (F::'a dom_F) (E::'a dom_F).
    (ALL (n::nat). F rest n = E rest n) --> F = E"
apply (intro allI impI)
apply (rule order_antisym)
apply (simp_all add: contra_diff_rs_dom_F_left)
done

(*** diff_rs_dom_F ***)

lemma diff_rs_dom_F: 
  "ALL (F::'a dom_F) (E::'a dom_F).
    F ~= E --> (EX (n::nat). F rest n ~= E rest n)"
apply (intro allI impI)
apply (erule contrapos_pp)
apply (insert contra_diff_rs_dom_F)
apply (drule_tac x="F" in spec)
apply (drule_tac x="E" in spec)
apply (simp)
done

(***************************************************************
                        dom_F ==> RS
 ***************************************************************)

instance dom_F :: (type) rs
apply (intro_classes)
apply (simp add: zero_eq_rs_dom_F)
apply (simp add: min_rs_dom_F)
apply (simp add: diff_rs_dom_F)
done

(************************************************************
                        dom_F ==> MS
 ************************************************************)

instance dom_F :: (type) ms
apply (intro_classes)
apply (simp add: positive_rs)
apply (simp add: diagonal_rs)
apply (simp add: symmetry_rs)
apply (simp add: triangle_inequality_rs)
done

(************************************************************
                 i.e.  dom_F ==> MS & RS 
 ************************************************************)

instance dom_F :: (type) ms_rs
apply (intro_classes)
done

(***********************************************************
                      lemmas (Limit)
 ***********************************************************)

(*** normal_seq lemma (Suc)***)

lemma normal_seq_dom_F_Suc:
  "ALL (Fs::'a dom_F infinite_seq) s X n. 
      normal Fs & Suc (lengtht s) <= n
         --> ((s,X) :f Fs (Suc (lengtht s))) = ((s,X) :f Fs n)"
apply (intro allI impI)
apply (erule conjE)+

apply (simp add: normal_def)
apply (drule_tac x="Suc (lengtht s)" in spec)
apply (drule_tac x="n" in spec)
apply (simp add: min_x_y)

apply (insert rest_nat_le_1_if)
apply (drule_tac x="Fs (Suc (lengtht s))" in spec)
apply (drule_tac x="Fs n" in spec)
apply (drule_tac x="Suc (lengtht s)" in spec)
apply (simp add: rest_F_def restFset_in Abs_dom_F_inject)
apply (simp add: restFset_def)

apply (simp add: Collect_open)
apply (drule_tac x="s" in spec)
apply (drule_tac x="X" in spec)
apply (simp)
done

(*** normal_seq lemma (<)***)

lemma normal_seq_dom_F_less:
  "ALL (Fs::'a dom_F infinite_seq) s X n. 
      normal Fs & lengtht s < n
         --> ((s,X) :f Fs (Suc (lengtht s))) = ((s,X) :f Fs n)"
apply (simp add: normal_seq_dom_F_Suc  less_nat_Suc_le)
done


(*** normal_seq lemma (Tick)***)

lemma normal_seq_dom_F_Tick:
  "ALL (Fs::'a dom_F infinite_seq) s' X n. 
      normal Fs & lengtht (s' @t [Tick]t) <= n & notick s'
         --> ((s' @t [Tick]t,X) :f Fs (lengtht (s' @t [Tick]t))) = 
             ((s' @t [Tick]t,X) :f Fs n)"
apply (intro allI impI)
apply (erule conjE)+

apply (simp add: normal_def)
apply (drule_tac x="lengtht (s' @t [Tick]t)" in spec)
apply (drule_tac x="n" in spec)
apply (simp add: min_x_y)

apply (insert rest_nat_le_1_if)
apply (drule_tac x="Fs (lengtht (s' @t [Tick]t))" in spec)
apply (drule_tac x="Fs n" in spec)
apply (drule_tac x="lengtht (s' @t [Tick]t)" in spec)
apply (simp add: rest_F_def restFset_in Abs_dom_F_inject)
apply (simp add: restFset_def)

apply (simp add: Collect_open)
apply (drule_tac x="s' @t [Tick]t" in spec)
apply (drule_tac x="X" in spec)
apply (simp)

apply (erule iffE)
apply (rule iffI)
apply (drule mp)
apply (simp)
apply (rule_tac x="s'" in exI)
apply (simp)
apply (simp)

apply (rotate_tac -2)
apply (drule mp)
apply (simp)
apply (rule_tac x="s'" in exI)
apply (simp)
apply (simp)
done

(*** limitFset_def in dom_F ***)

lemma limitFset_def_in:
  "ALL (Fs::'a dom_F infinite_seq).
    {sX. EX s. (EX X. sX = (s, X)) &
         (sX :f Fs (Suc (lengtht s)) |
          sX :f Fs (lengtht s) & (EX s'. s = s' @t [Tick]t & notick s'))}
      : dom_F"
apply (intro allI)
apply (simp add: dom_F_def CF1_def)
apply (intro allI impI)

apply (erule conjE)
apply (erule disjE)

apply (insert dom_F_F1)
apply (drule_tac x="s" in spec)
apply (drule_tac x="X" in spec)
apply (drule_tac x="Y" in spec)
apply (drule_tac x="Rep_dom_F (Fs (Suc (lengtht s)))" in spec)
apply (simp add: Rep_dom_F)

apply (drule_tac x="s" in spec)
apply (drule_tac x="X" in spec)
apply (drule_tac x="Y" in spec)
apply (drule_tac x="Rep_dom_F (Fs (lengtht s))" in spec)
apply (simp add: Rep_dom_F)
done

(*** limitFset in dom_F ***)

lemma limitFset_in:
  "ALL (Fs::'a dom_F infinite_seq). limitFset Fs : dom_F"
apply (simp add: limitFset_def_in limitFset_def)
done

(*** limit rest n ***)

lemma limitF_restFset:
  "ALL (Fs::'a dom_F infinite_seq) n.
        (limitF Fs) restFset n = 
        {sX. EX s. (EX X. sX = (s, X)) &
            (sX :f Fs (Suc (lengtht s)) |
             sX :f Fs (lengtht s) & (EX s'. s = s' @t [Tick]t & notick s')) &
            (lengtht s < n |
             lengtht s = n & (EX s'. s = s' @t [Tick]t & notick s'))}"
apply (intro allI)
apply (simp add: restFset_def)
apply (simp add: Collect_open)
apply (simp add: limitFset_in Abs_dom_F_inverse)

apply (intro allI)
apply (simp add: limitFset_def)
done

(*** limitF lemma ***)

lemma limitF_limit_lm:
  "ALL (Fs::'a dom_F infinite_seq). 
     normal Fs --> (ALL n. (limitF Fs) rest n = (Fs n) rest n)"
apply (intro allI impI)
apply (simp add: rest_F_def restFset_in Abs_dom_F_inject)

apply (simp add: limitF_restFset)
apply (simp add: restFset_def)

apply (simp add: Collect_open)
apply (intro allI)
apply (rename_tac Fs n s X)

apply (rule iffI)

(* => *)

apply (erule conjE)
apply (erule disjE)
apply (erule disjE)

apply (simp add: normal_seq_dom_F_less)

apply (erule conjE)
apply (erule exE)
apply (simp)
apply (insert normal_seq_dom_F_Tick)
apply (drule_tac x="Fs" in spec)
apply (drule_tac x="s'" in spec)
apply (drule_tac x="X" in spec)
apply (drule_tac x="Suc n" in spec)
apply (simp)
apply (rule_tac x="s'" in exI)
apply (simp)

apply (erule conjE)
apply (erule exE)
apply (simp)
apply (drule_tac x="Fs" in spec)
apply (drule_tac x="s'" in spec)
apply (drule_tac x="X" in spec)
apply (drule_tac x="n" in spec)
apply (erule disjE)
apply (simp)
apply (simp)

(* <= *)

apply (erule conjE)
apply (erule disjE)

apply (simp add: normal_seq_dom_F_less)
apply (auto)
done

(*** (normal) Fs converges to (limitF Fs) ***)

lemma limitF_limit:
  "ALL (Fs::'a dom_F infinite_seq). normal Fs --> Fs convergeTo (limitF Fs)"
apply (simp add: limitF_limit_lm rest_limit)
done

(*** (cauchy) Fs converges to (limitF NF Fs) ***)

lemma cauchy_limitF_limit:
  "ALL (Fs::'a dom_F infinite_seq). 
      cauchy Fs --> Fs convergeTo (limitF (NF Fs))"
apply (simp add: limitF_limit normal_form_seq_same_limit 
                 normal_form_seq_normal)
done

(************************************
    Dom_F --> Complete Metric Space
 ************************************)

lemma dom_F_cms:
  "ALL (Fs::'a dom_F infinite_seq). 
      cauchy Fs --> (EX F. Fs convergeTo F)"
apply (intro allI impI)
apply (rule_tac x="limitF (NF Fs)" in exI)
apply (simp add: cauchy_limitF_limit)
done

(************************************************************
                   dom_F ==> CMS and RS
 ************************************************************)

instance dom_F :: (type) cms
apply (intro_classes)
apply (simp add: dom_F_cms)
done

instance dom_F :: (type) cms_rs
apply (intro_classes)
done

(*** (normal) limit Fs = limitF Fs ***)

lemma limitF_limit_eq:
  "ALL (Fs::'a dom_F infinite_seq). normal Fs --> limit Fs = limitF Fs"
apply (intro allI impI)
apply (insert unique_convergence)
apply (drule_tac x="Fs" in spec)
apply (drule_tac x="limit Fs" in spec)
apply (drule_tac x="limitF Fs" in spec)
apply (simp add:  limitF_limit the_limit normal_cauchy)
done

(*----------------------------------------------------------*
 |                                                          |
 |         Fixed point induction for refinement             |
 |                                                          |
 *----------------------------------------------------------*)

consts
  Ref_F :: "'a dom_F => 'a dom_F => bool"
  Rev_F :: "'a dom_F => 'a dom_F => bool"

defs
  Ref_F_def : "Ref_F E == (%F. (E <= F))"
  Rev_F_def : "Rev_F E == (%F. (F <= E))"


(************************************************************
             continuity lemma for refinement
 ************************************************************)

lemma Ref_F_cont: "ALL E. continuous (Ref_F E)"
apply (intro allI)
apply (simp add: continuous_def)
apply (intro allI impI)
apply (simp add: Ref_F_def subset_F_def)
apply (simp add: subset_def Bex_def)
apply (elim conjE exE)
apply (rename_tac E F t X)

apply (rule_tac x="Suc (lengtht t)" in exI)
apply (intro allI impI)
apply (rule_tac x="t" in exI)
apply (rule_tac x="X" in exI)

apply (simp add: rest_F_def)
apply (simp add: restFset_in Abs_dom_F_inject)
apply (simp add: restFset_def)
apply (simp add: Collect_open)
apply (drule_tac x="t" in spec)
apply (drule_tac x="X" in spec)
apply (simp)
done

(************************************************************
         continuity lemma for (Reverse) refinement
 ************************************************************)

lemma Rev_F_cont: "ALL E. continuous (Rev_F E)"
apply (intro allI)
apply (simp add: continuous_def)
apply (intro allI impI)
apply (simp add: Rev_F_def subset_F_def)
apply (simp add: subset_def Bex_def)
apply (elim conjE exE)
apply (rename_tac E F t X)

apply (rule_tac x="Suc (lengtht t)" in exI)
apply (intro allI impI)
apply (rule_tac x="t" in exI)
apply (rule_tac x="X" in exI)

apply (simp add: rest_F_def)
apply (simp add: restFset_in Abs_dom_F_inject)
apply (simp add: restFset_def)
apply (simp add: Collect_open)
apply (drule_tac x="t" in spec)
apply (drule_tac x="X" in spec)
apply (simp)
done

(****************** to add them again ******************)

declare Union_image_eq [simp]
declare Inter_image_eq [simp]

end
