           (*-------------------------------------------*
            |                Csp-Prover                 |
            |                July 2004                  |
            |        Yoshinao Isobe (AIST JAPAN)        |
            *-------------------------------------------*)

theory Infra = Complex_Main: 

(*****************************************************
                       the
 *****************************************************)

lemma ALL_theI: "ALL P a. (P a & (ALL x. P x --> x=a)) --> P (THE x. P x)"
apply (auto simp add: theI)
done

(*****************************************************
                       fun
 *****************************************************)

consts
  exchange_args :: 
     "('x => ('y => 'z)) => 'y => ('x => 'z)" ("_<<_>>" [1000, 0] 1000)

defs
  exchange_args_def: "f<<y>> == (%x. ((f x) y))"

consts
  fst_fun  :: "('x => ('y * 'z)) => 'x => 'y"
  snd_fun  :: "('x => ('y * 'z)) => 'x => 'z"
  pair_rev :: "('x * 'y) => ('y * 'x)"
  pair_fun :: "('x => 'y) => ('x => 'z) => ('x => ('y * 'z))" 
                                           ("(2_ **/ _)" [51,52] 52)
  pair_fun_rev :: "('x => ('y * 'z)) => ('x => ('z * 'y))"

defs
  fst_fun_def  : "fst_fun f == (%x. (fst (f x)))"
  snd_fun_def  : "snd_fun f == (%x. (snd (f x)))"
  pair_rev_def : "pair_rev xc == (snd xc, fst xc)"
  pair_fun_def : "f ** g == (%x. (f x, g x))"
  pair_fun_rev_def : "pair_fun_rev fc == (%x. pair_rev (fc x))"

(* for convenience *)

lemma fun_neq: "ALL f g x. f x ~= g x --> f ~= g"
apply (auto)
done

lemma pair_fst_fun_rev:
      "ALL (fc::('x => ('a * 'b))). 
            (fst_fun (pair_fun_rev fc) = snd_fun fc)"
apply (simp add: fst_fun_def snd_fun_def pair_fun_rev_def pair_rev_def)
done

lemma pair_snd_fun_rev:
      "ALL (fc::('x => ('a * 'b))). 
            (snd_fun (pair_fun_rev fc) = fst_fun fc)"
apply (simp add: fst_fun_def snd_fun_def pair_fun_rev_def pair_rev_def)
done

lemma pair_fun_apply: 
  "ALL P (x::'a * 'b). ((%(y,z). P y z) x) = (P (fst x) (snd x))"
apply (auto)
done

(*****************************************************
                        List 
 *****************************************************)

lemma notin_butlast_app :
  "ALL e s t. e ~: set s & e ~: set (butlast t) -->
       e ~: set (butlast (s @ t))"
apply (intro allI)
apply (induct_tac s)
apply (auto)
done

lemma notin_butlast_app_inv :
  "ALL e s t. e ~: set (butlast (s @ t)) --> e ~: set (butlast s)"
apply (intro allI)
apply (induct_tac s)
apply (auto)
done

lemma notin_butlast_app_inv2 :
  "ALL e s t. e ~: set (butlast (s @ t)) --> e ~: set (butlast t)"
apply (intro allI)
apply (induct_tac s)
apply (auto)
done

lemma list_nil_or_unnil: "ALL t. (t = []) | (EX a s. t=a#s)"
apply (intro allI)
apply (induct_tac t)
apply (auto)
done

lemma list_last_nil_or_unnil: "ALL t. (t = []) | (EX s a. t= s@[a])"
apply (intro allI)
apply (induct_tac t)
apply (auto)
done

lemma butlast_subseteq : "ALL s. set (butlast s) <= set s"
apply (intro allI)
apply (induct_tac s)
apply (auto)
done

lemma non_nil_exists:
  "ALL t. t ~= [] --> (EX s a. t = s @ [a])"
apply (intro allI)
apply (induct_tac t)
apply (auto)
done

lemma ALL_butlast_snoc : "ALL t a. butlast (t @ [a]) = t"
apply (simp)
done

lemma app_list_length: "ALL s t. length s <= length (s @ t)"
apply (simp)
done

lemma one_list_decompo:
  "ALL a s t. [a] = s @ t --> s =[] | s=[a]"
apply (intro allI)
apply (induct_tac s)
apply (auto)
done

(*** length s = 1 ***)

lemma list_length_one: "ALL s. length s = Suc 0 --> (EX a. s =[a])"
apply (intro allI)
apply (induct_tac s)
apply (auto)
done

(*** list app app ***)

lemma list_app_app :
  "ALL s1 s2 t1 t2. s1 @ s2 = t1 @ t2 
       --> ((EX u. s1 = t1 @ u & t2 = u @ s2) |
            (EX u. t1 = s1 @ u & s2 = u @ t2)   )"
apply (rule allI)
apply (induct_tac s1)

apply (intro allI impI)
apply (rule disjI2)
apply (rule_tac x="t1" in exI)
apply (simp)

 (* step *)

apply (intro allI impI)

apply (insert list_nil_or_unnil)
apply (rotate_tac -1)
apply (drule_tac x="t1" in spec)
apply (erule disjE)

  (* t1 = [] *)
 apply (simp)

  (* t1 ~= [] *)
 apply (elim exE)
 apply (simp)
done

(*****************************************************
                       Set 
 *****************************************************)

consts
 open_uniset :: "'a set => 'a list"

defs
 open_uniset_def : 
    "open_uniset X == (if (EX a. X={a}) then [THE a. X={a}] else [])"

(* "nonempty_exE" and "ex_nonemptyE" add an assumption. *)

lemma nonempty_exE : 
  "[| X ~= {} ; [| X ~= {} ; EX x. x:X |] ==> R |] ==> R"
apply (auto)
done

lemma ex_nonemptyE : 
  "[| x:X ; [| X ~= {} ; x:X ; EX x. x:X|] ==> R |] ==> R"
apply (auto)
done

lemma ex_nonemptyI : "EX x. x:X ==> X ~= {}"
apply (auto)
done

lemma Collect_open: "ALL P Q. ({x. P(x)} = {x. Q(x)}) = (ALL x. P(x) = Q(x))"
apply (auto)
done

lemma in_image_set: "ALL x f X. x : f ` X --> (EX y. x = f y & y:X)"
apply (auto)
done

lemma Union_mono_all_ex: "ALL X Y. (ALL x : X. EX y : Y. x <= y)
  --> Union X <= Union Y"
apply (auto)
done

lemma image_Collect_fun:
 "ALL f g P. f ` {g x |x. P x} = {f (g x) |x. P x}"
apply (auto)
done

lemma Int_no_empty: "ALL a X Y. a : X & a : Y --> X Int Y ~= {}"
apply (auto)
done

lemma image_Un_dist:
   "ALL f X Y. f ` (X Un Y) = (f ` X) Un (f ` Y)"
apply (auto)
done

lemma image_Int_dist_inj:
   "ALL f X Y. inj f --> (f ` (X Int Y) = (f ` X) Int (f ` Y))"
apply (simp add: inj_on_def)
apply (auto)
done

lemma image_Minus_dist_inj:
   "ALL f X Y. inj f --> (f ` (X - Y) = (f ` X) - (f ` Y))"
apply (simp add: inj_on_def)
apply (auto)
done

lemma subsetE :
  "[| X <= Y ; [| ALL x:X. x:Y |] ==> R |] ==> R"
apply (auto)
done

(*****************************************************
                    order
 *****************************************************)

(*** ALL ***)

lemma ALL_order_less_le_trans:
    "ALL (x::'a::order) y z. ( x < y & y <= z ) --> x < z"
apply (auto)
done

(*****************************************************
                      nat
 *****************************************************)

lemma nat_pos_zero: "ALL n::nat. n=0 | (EX m::nat. n=m+1)"
apply (intro allI)
apply (induct_tac n)
apply (simp)
apply (rule disjI2)
apply (erule disjE)
apply (rule_tac x=0 in exI)
apply (simp)
apply (erule exE)
apply (rule_tac x="m+1" in exI)
apply (simp)
done

lemma ALL_nat_le_right_plus: "ALL (n::nat) m N. n <= m --> n + N <= m + N"
apply (auto)
done

lemma ALL_nat_le_diff_plus: 
  "ALL (n::nat) m N. (N <= n & n - N <= m) --> n <= m + N"
apply (intro allI impI)
apply (insert ALL_nat_le_right_plus)
apply (drule_tac x="n-N" in spec)
apply (drule_tac x="m" in spec)
apply (drule_tac x="N" in spec)
apply (simp)
done

lemma less_nat_Suc_le:
  "ALL (n::nat) (m::nat). n < m --> Suc n <= m"
apply (auto)
done

lemma Suc_le_nat_exists: 
  "ALL (M::nat) m. (Suc M <= m) --> (EX n. M <= n & m = Suc n)"
apply (intro allI)
apply (induct_tac m)
apply (auto)
done

(*****************************************************
                    Real number
 *****************************************************)

(*** ALL ***)

lemma ALL_reals_Archimedean2:"ALL (x::real). EX n::nat. x < real n"
apply (simp add: reals_Archimedean2)
done

lemma ALL_real_mult_left_mono: "ALL (a::real) (b::real) (c::real). 
   (a <= b & 0 <= c) --> c * a <= c * b"
apply (simp add: mult_left_mono)
done

lemma ALL_real_mult_right_mono: "ALL (a::real) (b::real) (c::real). 
   (a <= b & 0 <= c) --> a * c <= b * c"
apply (simp add: mult_right_mono)
done

lemma ALL_real_mult_le_cancel_if1: "ALL (x::real) (y::real) (z::real). 
   (x <= y & (0::real) < z) --> (x * z <= y * z)"
apply (simp)
done

lemma ALL_real_mult_le_cancel_imp1: "ALL (x::real) (y::real) (z::real). 
   (x * z <= y * z & (0::real) < z) --> (x <= y)"
apply (auto)
done

lemma ALL_real_mult_less_cancel_if1: "ALL (x::real) (y::real) (z::real). 
   (x < y & (0::real) < z) --> (x * z < y * z)"
apply (simp)
done

lemma ALL_real_mult_less_cancel_imp1: "ALL (x::real) (y::real) (z::real). 
   (x * z < y * z & (0::real) < z) --> (x < y)"
apply (auto)
done

lemma ALL_real_mult_less_cancel_if2: "ALL (x::real) (y::real) (z::real). 
   (x < y & (0::real) < z) --> (z * x < z * y)"
apply (simp add: real_mult_commute)
done

lemma ALL_real_mult_less_cancel_imp2: "ALL (x::real) (y::real) (z::real). 
   (z * x < z * y & (0::real) < z) --> (x < y)"
apply (auto simp add: real_mult_commute)
done

(*** derived lemma ***)

lemma ALL_div_mult_cancel: 
  "ALL (x::real) (y::real). y ~= 0 --> x / y * y = x"
apply (simp)
done

lemma ALL_real_add_mult_right_distrib:
  "ALL (x::real) (y::real) (z::real). (x + y) * z = x * z + y * z"
apply (simp add: real_add_mult_distrib)
done

lemma ALL_real_add_mult_left_distrib:
  "ALL (x::real) (y::real) (z::real). z * (x + y) = z * x + z * y"
apply (intro allI)
apply (insert ALL_real_add_mult_right_distrib)
apply (drule_tac x="x" in spec)
apply (drule_tac x="y" in spec)
apply (drule_tac x="z" in spec)
apply (simp add: real_mult_commute)
done

lemma ALL_real_diff_mult_right_distrib: 
  "ALL (x::real) (y::real) (z::real). (x - y) * z = x * z - y * z"
apply (simp add: real_diff_def ALL_real_add_mult_right_distrib)
done

lemma ALL_real_diff_mult_left_distrib: 
  "ALL (x::real) (y::real) (z::real). z * (x - y) = z * x - z * y"
apply (simp add: real_diff_def ALL_real_add_mult_left_distrib)
done

lemma ALL_real_divide_eq_eq:
   "ALL (x::real) (y::real) (z::real). 
       z ~= 0 --> (y / z = x) = (y = x * z)"
apply (intro allI impI)
apply (rule iffI)
apply (drule sym)
apply (simp_all)
done

lemma ALL_real_divide_le_eq:
   "ALL (x::real) (y::real) (z::real). 
       0 < z --> (x <= y / z) = (x * z <= y)"
apply (intro allI impI)
apply (rule iffI)
apply (insert ALL_real_mult_le_cancel_if1)
apply (drule_tac x="x" in spec)
apply (drule_tac x="y / z" in spec)
apply (drule_tac x="z" in spec)
apply (simp)

apply (simp)
apply (insert ALL_real_mult_le_cancel_imp1)
apply (drule_tac x="x" in spec)
apply (drule_tac x="y / z" in spec)
apply (drule_tac x="z" in spec)
apply (simp)
done

lemma ALL_real_le_divide_eq:
   "ALL (x::real) (y::real) (z::real). 
       0 < z --> (y / z <= x) = (y <= x * z)"
apply (intro allI impI)
apply (rule iffI)
apply (insert ALL_real_mult_le_cancel_if1)
apply (drule_tac x="y / z" in spec)
apply (drule_tac x="x" in spec)
apply (drule_tac x="z" in spec)
apply (simp)

apply (simp)
apply (insert ALL_real_mult_le_cancel_imp1)
apply (drule_tac x="y / z" in spec)
apply (drule_tac x="x" in spec)
apply (drule_tac x="z" in spec)
apply (simp)
done

lemma ALL_real_divide_less_eq:
   "ALL (x::real) (y::real) (z::real). 
       0 < z --> (x < y / z) = (x * z < y)"
apply (intro allI impI)
apply (rule iffI)
apply (insert ALL_real_mult_less_cancel_if1)
apply (drule_tac x="x" in spec)
apply (drule_tac x="y / z" in spec)
apply (drule_tac x="z" in spec)
apply (simp)

apply (simp)
apply (insert ALL_real_mult_less_cancel_imp1)
apply (drule_tac x="x" in spec)
apply (drule_tac x="y / z" in spec)
apply (drule_tac x="z" in spec)
apply (simp)
done

lemma ALL_real_less_divide_eq:
   "ALL (x::real) (y::real) (z::real). 
       0 < z --> (y / z < x) = (y < x * z)"
apply (intro allI impI)
apply (rule iffI)
apply (insert ALL_real_mult_less_cancel_if1)
apply (drule_tac x="y / z" in spec)
apply (drule_tac x="x" in spec)
apply (drule_tac x="z" in spec)
apply (simp)

apply (simp)
apply (insert ALL_real_mult_less_cancel_imp1)
apply (drule_tac x="y / z" in spec)
apply (drule_tac x="x" in spec)
apply (drule_tac x="z" in spec)
apply (simp)
done

lemma ALL_real_mult_order: 
  "ALL (x::real) (y::real). (0 < x & 0 < y ) --> (0::real) < x * y"
apply (simp add: real_mult_order)
done

lemma ALL_real_mult_order_zero: 
  "ALL (x::real) (y::real). (0 <= x & 0 <= y ) --> (0::real) <= x * y"
apply (intro allI impI)
apply (erule conjE)
apply (case_tac "x = 0", simp)
apply (case_tac "y = 0", simp)
apply (insert ALL_real_mult_order)
apply (drule_tac x="x" in spec)
apply (drule_tac x="y" in spec)
apply (auto)
done

lemma ALL_real_mult_div_commute: "ALL (x::real) (y::real) (z::real) (r::real).
         (0 <= x & 0 < y & 0 < z & 0 < r)
         --> (x < y * z / r) = (r * x / z < y)"
apply (intro allI impI)
apply (simp add: ALL_real_divide_less_eq)
apply (simp add: ALL_real_less_divide_eq)
apply (simp add: real_mult_commute)
done

lemma ALL_le_than_one: "ALL (a::real) (x::real) (y::real).
     (a < 1 & 0 <= y & x <= a * y) --> x <= y"
apply (intro allI impI)
apply (insert ALL_real_mult_right_mono)
apply (drule_tac x="a" in spec)
apply (drule_tac x="1" in spec)
apply (drule_tac x="y" in spec)
apply (auto)
done

(*****************************************************
            Min and Max for 2 (linorder)
 *****************************************************)

lemma max_x_y: "ALL (x::'a::linorder) (y::'a::linorder). x <= y --> max x y = y"
apply (simp add: max_def)
done

lemma max_y_x: "ALL (x::'a::linorder) (y::'a::linorder). x <= y --> max y x = y"
apply (auto simp add: max_def)
done

lemma min_x_y: "ALL (x::'a::linorder) (y::'a::linorder). x <= y --> min x y = x"
apply (simp add: min_def)
done

lemma min_y_x: "ALL (x::'a::linorder) (y::'a::linorder). y <= x --> min x y = y"
apply (auto simp add: min_def)
done

lemma x_max_x_y: "ALL (x::'a::linorder) (y::'a::linorder). x <= max x y"
apply (simp add: max_def)
done

lemma y_max_x_y: "ALL (x::'a::linorder) (y::'a::linorder). y <= max x y"
apply (auto simp add: max_def)
done

lemma x_min_x_y: "ALL (x::'a::linorder) (y::'a::linorder). min x y <= x"
apply (auto simp add: min_def)
done

lemma y_min_x_y: "ALL (x::'a::linorder) (y::'a::linorder). min x y <= y"
apply (simp add: min_def)
done

lemma greatest_min: "ALL x y z. z <= x & z <= y --> z <= min x y"
apply (simp add: min_def)
done

lemma least_max: "ALL x y z. x <= z & y <= z --> max x y <= z"
apply (simp add: max_def)
done

lemma real_max_plus: "ALL (x::real) (y::real). 
   (0 <= x & 0 <= y) --> max x y <= x + y"
apply (simp add: max_def)
done

lemma min_x_y_or: 
  "ALL (x::'a::linorder) (y::'a::linorder). min x y = x | min x y = y"
apply (intro allI)
apply (case_tac "x <= y")
apply (simp add: min_x_y)
apply (simp add: linorder_not_le)
apply (insert min_y_x)
apply (drule_tac x="x" in spec)
apply (drule_tac x="y" in spec)
apply (auto)
done

(*****************************************************
            Min and Max for predicates
 *****************************************************)

consts
  min_check::  "('a::linorder => bool) => 'a::linorder => bool"
  max_check::  "('a::linorder => bool) => 'a::linorder => bool"

defs
  min_check_def: "min_check P x == P x & (ALL y. P y --> x <= y)"
  max_check_def: "max_check P x == P x & (ALL y. P y --> y <= x)"


consts
  the_min::  "('a::linorder => bool) => 'a::linorder"
  minset ::  "('a::linorder set) => 'a::linorder"
  minset_check::  "('a::linorder set) => 'a::linorder => bool"

defs
  the_min_def : "the_min P == THE x. min_check P x"
  minset_def  : "minset X  == the_min (%x. x:X)"
  minset_check_def: "minset_check X x == min_check (%x. x:X) x"

(*** min lemma ***)

lemma exists_min_nat_lm:
  "ALL P (n::nat). (EX n'. (P n' & n' <= n)) --> (EX m. min_check P m)"
apply (intro allI)
apply (induct_tac n)
apply (simp)
apply (intro impI)
apply (rule_tac x="0" in exI)
apply (simp add: min_check_def)

apply (intro impI)
apply (erule exE)

apply (case_tac "n' <= na")
 apply (drule mp)
 apply (rule_tac x="n'" in exI)
 apply (simp)
 apply (simp)

apply (case_tac "n' = Suc na")
 apply (case_tac "EX n'. P n' & n' <= na")
  apply (simp)

  apply (rule_tac x="Suc na" in exI)
  apply (simp add: min_check_def)
  apply (intro allI impI)
  apply (drule_tac x="y" in spec)
  apply (simp)

(* else --> contradiction *)
apply (simp)
done

(*** minimum existency ***)

lemma exists_min_nat:
  "ALL P (n::nat). P n --> (EX m. min_check P m)"
apply (intro allI impI)
apply (insert exists_min_nat_lm)
apply (drule_tac x="P" in spec)
apply (drule_tac x="n" in spec)
apply (drule mp)
apply (rule_tac x="n" in exI)
apply (simp_all)
done

(*** minimum unique ***)

lemma unique_min_nat:
  "ALL P (m::nat) (n::nat). ( min_check P m & min_check P n) --> m=n"
apply (intro allI impI)
apply (erule conjE)
apply (simp add: min_check_def)
apply (elim conjE)
apply (drule_tac x="n" in spec)
apply (drule_tac x="m" in spec)
apply (simp)
done

(*** maximum unique ***)

lemma unique_max_nat:
  "ALL P (m::nat) (n::nat). ( max_check P m & max_check P n) --> m=n"
apply (intro allI impI)
apply (erule conjE)
apply (simp add: max_check_def)
apply (elim conjE)
apply (drule_tac x="n" in spec)
apply (drule_tac x="m" in spec)
apply (simp)
done

(*** the_min ***)

lemma the_min_check:
  "ALL (P::nat => bool). (EX a. P a) --> min_check P (the_min P)"
apply (intro allI impI)
apply (erule exE)

apply (insert exists_min_nat)
apply (drule_tac x="P" in spec)
apply (drule_tac x="a" in spec)
apply (simp)
apply (erule exE)

apply (insert unique_min_nat)
apply (drule_tac x="P" in spec)
apply (drule_tac x="x" in spec)

apply (simp add: the_min_def)
apply (insert ALL_theI)
apply (drule_tac x="min_check P" in spec)
apply (rotate_tac -1)
apply (drule_tac x="x" in spec)
apply (drule mp)

apply (rule conjI)
apply (assumption)
apply (intro allI)
apply (drule_tac x="xa" in spec)
apply (simp)
apply (simp)
done

(*** minset ***)

lemma the_minset_check:
  "ALL (X::nat set). X ~={} --> minset_check X (minset X)"
apply (intro allI impI)
apply (simp add: minset_def minset_check_def)
apply (erule nonempty_exE)
apply (auto simp add: the_min_check)
done


(*****************************************************
            Upper Bound and Lower Bound
 *****************************************************)

consts
  is_ub  :: "'a::order set => 'a::order => bool"  ("_ <| _" [55,55] 55)
  is_lub :: "'a::order set => 'a::order => bool"  ("_ <<| _"[55,55] 55)
  is_lb  :: "'a::order => 'a::order set => bool"  ("_ |< _" [55,55] 55)
  is_glb :: "'a::order => 'a::order set => bool"  ("_ |<< _"[55,55] 55)

  sup    :: "'a::order set => 'a::order"
  inf    :: "'a::order set => 'a::order"

defs
  is_ub_def  : "X <| x  == ALL y. y:X --> y <= x"
  is_lub_def : "X <<| x == (X <| x) & (ALL y. X <| y --> x <= y)"
  is_lb_def  : "x |< X  == ALL y. y:X --> x <= y"
  is_glb_def : "x |<< X == (x |< X) & (ALL y. y |< X --> y <= x)"

  sup        : "sup X == THE x. (X <<| x)"
  inf        : "inf X == THE x. (x |<< X)"

(*** LUB is unique ***)

lemma lub_unique : "ALL X x y. (X <<| x & X <<| y ) --> x = y"
apply (intro allI impI)
apply (unfold is_lub_def)
apply (elim conjE)
apply (drule_tac x="y" in spec)
apply (drule_tac x="x" in spec)
apply (simp)
done

(*** GLB is unique ***)

lemma glb_unique : "ALL X x y. (x |<< X & y |<< X ) --> x = y"
apply (intro allI impI)
apply (unfold is_glb_def)
apply (elim conjE)
apply (drule_tac x="y" in spec)
apply (drule_tac x="x" in spec)
apply (simp)
done

(*****************************************************
                  Power function
 *****************************************************)

(*** ALL ***)

lemma ALL_powr_log_cancel: 
  "ALL (a::real) (x::real). (0 < a & a ~= 1 & 0 < x )
       --> a powr (log a x) = x"
apply (simp)
done

(* derived lemmas *)

lemma ALL_realpow_le_le: (* power_decreasing *)
  "ALL (a::real) (m::nat) (n::nat).
       (0 <= a & a <= (1::real) & m <= n) --> a^n <= a^m"
apply (simp add: power_decreasing)
done

lemma ALL_realpow_Suc_less:
  "ALL (r::real) (n::nat).
     (0 < r & r < 1) --> r ^ Suc n < r ^ n"
apply (intro allI)
apply (induct_tac n)
apply (auto simp add: ALL_real_mult_less_cancel_if2)
done

lemma ALL_realpow_le_le_inv:
  "ALL (r::real) (m::nat) (n::nat).
     (0 < r & r < 1 & r ^ n <= r ^ m) -->  m <= n"
apply (intro allI impI)
apply (elim conjE)

apply (case_tac "m <= n")
 apply (simp)

(* else n < m --> contradiction *)

 apply (insert ALL_realpow_le_le)
 apply (drule_tac x="r" in spec)
 apply (drule_tac x="Suc n" in spec)
 apply (drule_tac x="m" in spec)

 apply (insert ALL_realpow_Suc_less)
 apply (drule_tac x="r" in spec)
 apply (drule_tac x="n" in spec)
 apply (simp)
done

lemma ALL_realpow_less_less_inv:
  "ALL (r::real) (m::nat) (n::nat).
     (0 < r & r < 1 & r ^ n < r ^ m) -->  m < n"
apply (intro allI impI)
apply (elim conjE)
apply (case_tac "n <= m")

apply (insert ALL_realpow_le_le)
apply (drule_tac x="r" in spec)
apply (drule_tac x="n" in spec)
apply (drule_tac x="m" in spec)
apply (auto)
done


lemma pow_min_max_dist:
  "ALL (r::real) (n::nat) (m::nat).
     (0 < r & r < 1) --> (r ^ (min n m)) = max (r^n) (r^m)"
apply (intro allI)
apply (simp add: max_def)
apply (rule conjI)
apply (intro impI)
apply (insert ALL_realpow_le_le_inv)
apply (drule_tac x="r" in spec)
apply (drule_tac x="m" in spec)
apply (drule_tac x="n" in spec)
apply (simp add: min_y_x)

apply (intro impI)
apply (drule_tac x="r" in spec)
apply (drule_tac x="n" in spec)
apply (drule_tac x="m" in spec)
apply (simp add: min_x_y)
done

lemma ALL_realpow_le_Suc_less:
  "ALL (x::real) (r::real) (n::nat). 
     (0 < r & r < 1 & x <= r^(Suc n)) --> x < r^n"
apply (intro allI impI)
apply (insert ALL_realpow_Suc_less)
apply (drule_tac x="r" in spec)
apply (drule_tac x="n" in spec)
apply (auto)
done

(*****************************
         powr --> pow  
 *****************************)

lemma nat_powr_pow:
  "ALL (r::real) (n::nat). 0 < r --> r powr (real n) = r ^ n"
apply (intro allI)
apply (induct_tac n)
apply (simp)
apply (simp add: real_of_nat_Suc)
apply (simp add: powr_add)
done

lemma ALL_powr_add:
  "ALL (x::real) (y::real) (z::real). 
      x powr (y + z) = (x powr y) * (x powr z)"
apply (simp add: powr_add)
done

lemma ALL_pow_add:
  "ALL (r::real) (m::nat) (n::nat). 
     0 < r --> r^(n + m) = (r^n) * (r^m)"
apply (intro allI impI)
apply (insert nat_powr_pow)
apply (frule_tac x="r" in spec)
apply (frule_tac x="r" in spec)
apply (drule_tac x="r" in spec)
apply (drule_tac x="n + m" in spec)
apply (drule_tac x="n" in spec)
apply (drule_tac x="m" in spec)
apply (simp add: ALL_powr_add)
done

(*****************************************************
              Exponentail convergence
 *****************************************************)

lemma powr_less_mono_inv:
  "ALL (a::real) (x::real) (y::real).(1 < a & x<y) 
        --> (inverse a) powr y < (inverse a) powr x"
apply (simp add: powr_def)
apply (auto simp add: ln_inverse)
done

lemma powr_less_mono_conv:
  "ALL (a::real) (x::real) (y::real).(0 < a & a < 1 & x < y)
        --> a powr y < a powr x"
apply (intro allI impI)
apply (insert powr_less_mono_inv)
apply (drule_tac x="inverse a" in spec)
apply (drule_tac x="x" in spec)
apply (drule_tac x="y" in spec)
apply (simp)
apply (drule mp)
apply (rule inverse_less_imp_less)
apply (simp_all)
done

lemma powr_convergence:
  "ALL (alpha::real) (x::real).
    ( 0 < alpha & alpha < 1 & 0 < x )
      --> (EX n::nat. alpha powr (real n) < x)"
apply (intro allI impI)

apply (insert ALL_powr_log_cancel)
apply (drule_tac x="alpha" in spec)
apply (drule_tac x="x" in spec)

apply (insert ALL_reals_Archimedean2)
apply (drule_tac x="log alpha x" in spec)
apply (erule exE)

apply (insert powr_less_mono_conv)
apply (drule_tac x="alpha" in spec)
apply (drule_tac x="log alpha x" in spec)
apply (drule_tac x="real n" in spec)

apply (rule_tac x="n" in exI)
apply (auto)
done

lemma pow_convergence:
  "ALL (alpha::real) (x::real).
    (0 <= alpha & alpha < 1 & 0 < x ) --> (EX n::nat. alpha^n < x)"
apply (intro allI impI)

apply (case_tac "alpha=0")
 apply (rule_tac x="1" in exI)
 apply (simp)

apply (case_tac "0<alpha")
 apply (insert powr_convergence)
 apply (drule_tac x="alpha" in spec)
 apply (drule_tac x="x" in spec)
 apply (simp)
 apply (erule exE)
 apply (rule_tac x="n" in exI)
 apply (simp add: nat_powr_pow)

apply (simp add: real_less_le)
done

(*** if 0 <=alpha < 1 then alpha^n ----> 0 ***)

lemma pow_glb:
  "ALL (alpha::real) (x::real).
    (0 <= alpha & alpha < 1) --> (0 |<< {r. EX n. r = alpha ^ n})"
apply (intro allI impI)
apply (simp add: is_glb_def is_lb_def)
apply (rule conjI)

(* lb *)

apply (intro allI impI)
apply (erule exE)
apply (simp add: zero_le_power)

(* glb *)

apply (intro allI impI)

apply (case_tac "y <=0")
apply (simp)

 (* 0 < y *)

 apply (insert pow_convergence)
 apply (rotate_tac -1)
 apply (drule_tac x="alpha" in spec)
 apply (rotate_tac -1)
 apply (drule_tac x="y" in spec)

 apply (simp)
 apply (erule exE)
 apply (drule_tac x="alpha ^ n" in spec)
 apply (drule mp)
 apply (rule_tac x="n" in exI)
 apply (simp_all)

done

(*****************************************************
                    Progression
 *****************************************************)

consts
  prog_sum0 :: "nat => (nat => real) => real"
  prog_sum  :: "nat => nat => (nat => real) => real"

primrec
  "prog_sum0 (Suc n) f = (prog_sum0 n f)+f(Suc n)"
  "prog_sum0 0 f = 0"

defs
  prog_sum_def : 
    "prog_sum m n f == (prog_sum0 n f) - (prog_sum0 m f)"


lemma geo_prog_sum0: 
  "ALL n (K::real) (alpha::real). 
     prog_sum0 n (%n::nat. (alpha^(n-(Suc 0))) * K) *((1::real)-alpha)
       = K*((1::real)-(alpha^n))"
apply (intro allI)
apply (induct_tac n)
apply (simp)
apply (simp add: real_add_mult_distrib)
apply (simp add: ALL_real_diff_mult_left_distrib)
done

lemma geo_prog_sum: 
  "ALL m n (K::real) (alpha::real). 
     prog_sum m n (%n::nat. (alpha^(n- (Suc 0)))*K) *((1::real)-alpha)
       = K*((alpha^m)-(alpha^n))"
apply (unfold prog_sum_def)
apply (simp only: ALL_real_diff_mult_right_distrib geo_prog_sum0)
apply (simp add: ALL_real_diff_mult_left_distrib)
done

lemma geo_prog_sum_div: 
  "ALL m n (K::real) (alpha::real). 
     alpha < (1::real) -->
     prog_sum m n (%n::nat. (alpha^(n-(Suc 0)))*K)
       = K*((alpha^m)-(alpha^n))/((1::real)-alpha)"
apply (intro allI impI)
apply (rule sym)
apply (simp add: ALL_real_divide_eq_eq)
apply (simp add: geo_prog_sum)
done


lemma geo_prog_sum_infinite: 
  "ALL m n (K::real) (alpha::real). 
     ((0::real) <= K & (0::real) <= alpha) -->
     prog_sum m n (%n::nat. (alpha^(n- (Suc 0)))*K) *((1::real)-alpha)
       <= K*(alpha^m)"
apply (intro allI impI)
apply (simp add: geo_prog_sum)
apply (simp add: ALL_real_diff_mult_left_distrib)
apply (simp add: zero_le_power ALL_real_mult_order_zero)
done

lemma geo_prog_sum_infinite_div: 
  "ALL m n (K::real) (alpha::real). 
     ((0::real) <= K & (0::real) <= alpha & alpha < (1::real)) -->
     prog_sum m n (%n::nat. (alpha^(n- (Suc 0)))*K)
       <= K*(alpha^m)/((1::real)-alpha)"
apply (simp add: ALL_real_divide_le_eq geo_prog_sum_infinite)
done

(*****************************************************
                  Infinite Sequence
 *****************************************************)

types 'a infinite_seq = "nat => 'a"           (* synonym *)

(*****************************************************
                         Pair
 *****************************************************)

lemma eq_pair_decompo:
  "ALL (xc::('x * 'y)) (yc::('x * 'y)).
      (xc = yc) = ((fst xc = fst yc) & (snd xc = snd yc))"
apply (simp)
done

lemma neq_pair_decompo:
  "ALL (xc::('x * 'y)) (yc::('x * 'y)).
      (xc ~= yc) = ((fst xc ~= fst yc) | (snd xc ~= snd yc))"
apply (simp)
done

lemma pair_fst_set_nonemp: "ALL sc. sc ~= {} --> fst ` sc ~= {}"
apply (simp)
done

lemma pair_snd_set_nonemp: "ALL sc. sc ~= {} --> snd ` sc ~= {}"
apply (simp)
done

lemma pair_fst_set_nonempE: 
  "[| sc ~= {} ; [| sc ~= {} ; fst ` sc ~= {} |] ==> R |] ==> R"
apply (simp)
done

lemma pair_snd_set_nonempE: 
  "[| sc ~= {} ; [| sc ~= {} ; snd ` sc ~= {} |] ==> R |] ==> R"
apply (simp)
done

lemma pair_fst_set_ex: 
  "ALL x sc. x : fst ` sc --> (EX y. y : snd ` sc & (x, y):sc)"
apply (auto)
done

lemma pair_snd_set_ex: 
  "ALL y sc. y : snd ` sc --> (EX x. x : fst ` sc & (x, y):sc)"
apply (auto)
done

(*** rev ***)

lemma pair_rev_eq:
      "ALL (xc::('a * 'b)) (yc::('a * 'b)).
          (pair_rev xc = pair_rev yc) =(xc = yc)"
apply (intro allI)
apply (simp add: pair_rev_def)
apply (simp add: Pair_fst_snd_eq)
apply (auto)
done

lemma pair_rev_neq:
      "ALL (xc::('a * 'b)) (yc::('a * 'b)).
           (pair_rev xc ~= pair_rev yc) = (xc ~= yc)"
apply (simp add: pair_rev_eq)
done

(*******************************
         <= in pair
 *******************************)

instance * :: (order,order) ord 
apply (intro_classes)
done

defs (overloaded)
  order_pair_def:
     "xc <= yc  ==  (fst xc <= fst yc) & (snd xc <= snd yc)"
  order_less_pair_def:
     "xc <  yc  ==  (fst xc <= fst yc) & (snd xc <=  snd yc) &
                   ((fst xc ~= fst yc) | (snd xc ~=  snd yc))"

(*** order in pair ***)

instance * :: (order,order) order
apply (intro_classes)
apply (unfold order_pair_def order_less_pair_def)
apply (simp)
apply (blast intro: order_trans)

apply (simp add: eq_pair_decompo)
apply (blast intro: order_antisym)

apply (auto simp add: neq_pair_decompo)
done

(*****************************************************
                      fun (Product)
 *****************************************************)

(*******************************
        <= in fun
 *******************************)

instance fun :: (type,order) ord 
apply (intro_classes)
done

defs (overloaded)
  order_prod_def:
     "xp <= yp  ==  ALL x. (xp x <= yp x)"
  order_less_prod_def:
     "xp < yp  ==  (ALL x. (xp x <= yp x)) &
                   (EX  x. (xp x ~= yp x))"

(*** order in prod ***)

instance fun :: (type,order) order
apply (intro_classes)
apply (unfold order_prod_def order_less_prod_def)
apply (simp)
apply (blast intro: order_trans)

apply (simp add: expand_fun_eq)
apply (blast intro: order_antisym)

apply (auto simp add: expand_fun_eq)
done

lemma order_prod_inv: "ALL f g. (ALL x. f x <= g x) --> f <= g"
apply (simp add: order_prod_def)
done

(*****************************************************
                     Fix Point
 *****************************************************)

consts
 is_ufp :: "'a::order => ('a::order => 'a::order) => bool" (infixl "isUFP" 55) 
 is_lfp :: "'a::order => ('a::order => 'a::order) => bool" (infixl "isLFP" 55)
 is_gfp :: "'a::order => ('a::order => 'a::order) => bool" (infixl "isGFP" 55)

  UFP    :: "('a::order => 'a::order) => 'a::order"
  LFP    :: "('a::order => 'a::order) => 'a::order"
  GFP    :: "('a::order => 'a::order) => 'a::order"

defs
  is_ufp_def : "x isUFP f == (x = f x & (ALL y. y = f y --> x  = y))"
  is_lfp_def : "x isLFP f == (x = f x & (ALL y. y = f y --> x <= y))"
  is_gfp_def : "x isGFP f == (x = f x & (ALL y. y = f y --> y <= x))"

  UFP_def    : "UFP f    == THE x. x isUFP f"
  LFP_def    : "LFP f    == THE x. x isLFP f"
  GFP_def    : "GFP f    == THE x. x isGFP f"

(*******************************
            lemmas 
 *******************************)

(*** UFP is unique ***)

lemma ufp_unique : "ALL f x y. x isUFP f & y isUFP f --> x = y"
apply (simp add: is_ufp_def)
done

(*** LFP is unique ***)

lemma lfp_unique : "ALL f x y. x isLFP f & y isLFP f --> x = y"
apply (intro allI)
apply (intro impI)
apply (unfold is_lfp_def)
apply (erule conjE)+
apply (drule_tac x="y" in spec)
apply (drule_tac x="x" in spec)
apply (simp)
done

lemma lfp_uniqueI : "[| x isLFP f ; y isLFP f |] ==> x = y"
apply (unfold is_lfp_def)
apply (erule conjE)+
apply (drule_tac x="y" in spec)
apply (drule_tac x="x" in spec)
apply (simp)
done

(*** GFP is unique ***)

lemma gfp_unique : "ALL f x y. x isGFP f & y isGFP f --> x = y"
apply (intro allI impI)
apply (unfold is_gfp_def)
apply (erule conjE)+
apply (drule_tac x="y" in spec)
apply (drule_tac x="x" in spec)
apply (simp)
done

lemma gfp_uniqueI : "[| x isGFP f ; y isGFP f |] ==> x = y"
apply (unfold is_gfp_def)
apply (erule conjE)+
apply (drule_tac x="y" in spec)
apply (drule_tac x="x" in spec)
apply (simp)
done

(*** THE UFP ***)

lemma the_ufp : "ALL x f. x isUFP f --> UFP f = x"
apply (intro allI impI)
apply (simp add: UFP_def)
apply (rule the_equality)
apply (simp)
apply (simp add: is_ufp_def)
done

(*** THE LFP ***)

lemma the_lfp : "ALL x f. x isLFP f --> LFP f = x"
apply (intro allI impI)
apply (simp add: LFP_def)
apply (rule the_equality)
apply (simp)
apply (rule lfp_uniqueI, simp_all)
done

(*** THE GFP ***)

lemma the_gfp : "ALL x f. x isGFP f --> GFP f = x"
apply (intro allI impI)
apply (simp add: GFP_def)
apply (rule the_equality)
apply (simp)
apply (rule gfp_uniqueI, simp_all)
done

end
