           (*-------------------------------------------*
            |                Csp-Prover                 |
            |                June 2004                  |
            |        Yoshinao Isobe (AIST JAPAN)        |
            *-------------------------------------------*)

theory Prefix = Trace:

(*  The following simplification rules are deleted in this theory file *)
(*  because they unexpectly rewrite (notick | t = []t)                 *)
(*                                                                     *)
(*                  disj_not1: (~ P | Q) = (P --> Q)                   *)

declare disj_not1 [simp del]

(*********************************************************)

consts
  prefix_list    :: "'a event list => 'a event list => bool"

defs
  prefix_list_def : 
    "prefix_list s t == EX u. (t = s @ u & t : trace)"

syntax
  "_prefix" :: "'a trace => 'a trace => bool" ("prefix")

translations
  "prefix s t" == "prefix_list (Rep_trace s) (Rep_trace t)"

consts
  prefix_closed  :: "('a trace set) => bool"

defs
  prefix_closed_def : 
    "prefix_closed T == ALL s t. ((t : T & prefix s t) --> s : T)"

(***********************************************************
                       lemmas (prefix)
 ***********************************************************)

(*** another def ***)

lemma prefix_iff: 
  "ALL s t. prefix s t = (EX u. (Rep_trace t = Rep_trace s @ u))"
apply (simp add: prefix_list_def)
apply (simp add: Rep_trace)
done

(*** Prefix app ***)

lemma prefix_app: "ALL s t. notick s --> prefix s (s @t t)"
apply (intro allI impI)
apply (simp add: prefix_list_def)
apply (simp add: Rep_trace)
apply (rule_tac x="Rep_trace t" in exI)

apply (insert trace_app_in_trace_if)
apply (drule_tac x="s" in spec)
apply (drule_tac x="t" in spec)
apply (simp add: Abs_trace_inverse)
done

(*** [] is a prefix of any trace ***)

lemma nil_is_prefix_list: "ALL t. t:trace --> prefix_list [] t"
apply (simp add: prefix_list_def)
done

lemma nil_t_is_prefix: "ALL t. prefix []t t"
apply (simp add: prefix_iff)
apply (simp add: Abs_trace_inverse trace_rep_nil_t)
done

(*** the prefix of []t is []t ***)

lemma prefix_list_of_nil: "ALL s. prefix_list s [] --> s = []"
apply (simp add: prefix_list_def)
done

lemma prefix_of_nil_t: "ALL s. prefix s []t --> s = []t"
apply (simp add: prefix_iff)
apply (simp add: nil_in_trace Abs_trace_inverse Abs_Rep_t_inv)
done

(*** the prefix of [a] is [] or [a] (list) ***)

lemma prefix_list_of_one: "ALL s a. prefix_list s [a] --> s = [] | s = [a]"
apply (intro allI impI)
apply (simp add: prefix_list_def)
apply (simp add: event_in_trace)
apply (erule exE)
apply (insert one_list_decompo)
apply (drule_tac x="a" in spec)
apply (drule_tac x="s" in spec)
apply (drule_tac x="u" in spec)
apply (simp)
done

lemma prefix_of_one: "ALL s a. prefix s [a]t --> s = []t | s = [a]t"
apply (intro allI impI)
apply (insert prefix_list_of_one)
apply (drule_tac x="Rep_trace s" in spec)
apply (drule_tac x="a" in spec)
apply (simp add: event_in_trace nil_in_trace Abs_trace_inverse Abs_Rep_t_inv)
done

(*** length of prefix ***)

lemma prefix_trace_length:
  "ALL s t. prefix s t --> lengtht s <= lengtht t"
apply (intro allI impI)
apply (simp add: prefix_iff)
apply (erule exE)
apply (simp)
done

(*** to avoide infinite rewriting ***)

lemma ALL_prefix_closed:
  "ALL T. prefix_closed T --> (ALL s t. ((t : T & prefix s t) --> s : T))"
apply (intro allI impI)
apply (unfold prefix_closed_def)
apply (blast)
done

(*** eliminate "prefix_closed" in assumptions by "erule prefix_closedE" ***)

lemma prefix_closedE : 
  "[| prefix_closed T ; ALL s t. ((t : T & prefix s t) --> s : T) ==> R |] 
     ==> R"
apply (unfold prefix_closed_def)
apply (blast)
done

(*** prefix closed & prefix ***)

lemma prefix_closed_propI:
  "[| prefix_closed T ; t : T ; prefix s t |] ==> s : T"
apply (unfold prefix_closed_def)
apply (blast)
done

(***********************************************************
                   convenient lemmas
 ***********************************************************)

(*** prefix a#v u ***)

lemma prefix_list_head_same: 
  "ALL a v u. prefix_list (a#v) u --> (EX u'. u = a#u' & prefix_list v u')"
apply (intro allI impI)
apply (simp add: prefix_list_def)

apply (erule conjE)
apply (erule exE)
apply (rule_tac x="v @ ua" in exI)
apply (simp)
apply (insert event_app_in_trace_inv)
apply (drule_tac x="a" in spec)
apply (drule_tac x="v @ ua" in spec)
apply (simp)
done

(*** prefix v a#u ***)

lemma prefix_list_head_same_inv: 
  "ALL v a u. prefix_list v (a#u)
      --> v = [] | (EX v'. v = a#v' & prefix_list v' u)"
apply (intro allI impI)
apply (simp add: prefix_list_def)

apply (erule conjE)
apply (erule exE)
apply (insert list_nil_or_unnil)
apply (drule_tac x="v" in spec)
apply (erule disjE)
apply (simp)

apply (erule exE)+
apply (simp)
apply (blast intro: event_app_in_trace_invI)
done

(*** prefix_list s t --> s:trace ***)

lemma prefix_list_s_trace: "ALL s t. prefix_list s t --> s:trace"
apply (intro allI impI)
apply (simp add: prefix_list_def)
apply (erule conjE)
apply (erule exE)

apply (insert list_last_nil_or_unnil)
apply (drule_tac x="u" in spec)
apply (erule disjE)
apply (simp)

apply (erule exE)+
apply (simp add: trace_def)

apply (insert ALL_butlast_snoc)
apply (drule_tac x="s @ sa" in spec)
apply (drule_tac x="a" in spec)
apply (simp)
apply (insert butlast_subseteq)
apply (drule_tac x="s" in spec)
apply (blast)
done

(****************** to add it again ******************)


declare disj_not1 [simp]

end
