           (*-------------------------------------------*
            |                Csp-Prover                 |
            |                June 2004                  |
            |        Yoshinao Isobe (AIST JAPAN)        |
            *-------------------------------------------*)

theory RS_pair = RS:

(*****************************************************************

         1. Pairs of RS are also RS.
         2. 
         3. 
         4. 

 *****************************************************************)

(********************************************************** 
              def: pair of restriction space
 **********************************************************)

instance * :: (rs,rs) ms0
apply (intro_classes)
done

defs (overloaded)
  pair_restriction_def : 
    "(xc::('a::rs * 'b::rs)) rest n 
        == (((fst xc) rest n) , ((snd xc) rest n))"

consts
  pair_limit :: "(('a::ms0 * 'b::ms0) infinite_seq) => ('a::ms0 * 'b::ms0)"

defs
  pair_limit_def :
     "pair_limit xcs == (limit (fst_fun xcs) , limit (snd_fun xcs))"


(************************************************************
                      Basics
 ************************************************************)

lemma rest_to_pair_rest:
    "ALL (xc::('a::rs * 'b::rs)) (yc::('a::rs * 'b::rs)) n.
         ((fst xc) rest n = (fst yc) rest n &
          (snd xc) rest n = (snd yc) rest n)
              --> xc rest n = yc rest n"
apply (simp add: pair_restriction_def)
done

(************************************************************
                   Pair RS ==> RS 
 ************************************************************)

(*** pair_zero_eq_rs ***)

lemma pair_zero_eq_rs:
  "ALL (xc::('a::rs * 'b::rs)) (yc::('a::rs * 'b::rs)).
     xc rest 0 = yc rest 0"
apply (intro allI)
apply (simp add: pair_restriction_def)
apply (simp add: zero_eq_rs)
done

(*** pair_min_rs ***)

lemma pair_min_rs: 
  "ALL (xc::('a::rs * 'b::rs)) (m::nat) (n::nat). 
           (xc rest m) rest n = xc rest (min m n)"
apply (intro allI)
apply (simp add: pair_restriction_def)
apply (simp add: min_rs)
done

(*** contra_pair_diff_rs ***)

lemma contra_pair_diff_rs:
  "ALL (xc::('a::rs * 'b::rs)) (yc::('a::rs * 'b::rs)).
                (ALL n. xc rest n = yc rest n) --> (xc = yc)"
apply (intro allI impI)
apply (simp add: pair_restriction_def)

apply (insert contra_diff_rs)
apply (drule_tac x="fst xc" in spec)
apply (drule_tac x="fst yc" in spec)
apply (insert contra_diff_rs)
apply (drule_tac x="snd xc" in spec)
apply (drule_tac x="snd yc" in spec)

apply (simp add: Pair_fst_snd_eq)
done

(*** pair_diff_rs ***)

lemma pair_diff_rs: 
  "ALL (xc::('a::rs * 'b::rs)) (yc::('a::rs * 'b::rs)).
                (xc ~= yc) --> (EX n. xc rest n ~= yc rest n)"
apply (intro allI impI)
apply (erule contrapos_pp)
apply (insert contra_pair_diff_rs)
apply (drule_tac x="xc" in spec)
apply (drule_tac x="yc" in spec)
apply (simp)
done

(*****************************************
              Pair RS => RS
 *****************************************)

instance * :: (rs,rs) rs
apply (intro_classes)
apply (simp add: pair_zero_eq_rs)
apply (simp add: pair_min_rs)
apply (simp add: pair_diff_rs)
done

(************************************************************
                     Pair RS ==> MS 
 ************************************************************)

instance * :: (rs,rs) ms
apply (intro_classes)
apply (simp add: positive_rs)
apply (simp add: diagonal_rs)
apply (simp add: symmetry_rs)
apply (simp add: triangle_inequality_rs)
done

(************************************************************
             i.e.  Pair RS ==> MS & RS 
 ************************************************************)

instance * :: (rs,rs) ms_rs
apply (intro_classes)
done

(************************************************************
                     Lemmas (distance)
 ************************************************************)

(*** distance pair_rev = distance pair ***)

lemma pair_rev_distance_eq:
      "ALL (xc::('a::rs * 'b::rs)) (yc::('a::rs * 'b::rs)).
          distance(pair_rev xc, pair_rev yc) = distance(xc, yc)"
apply (intro allI)

apply (case_tac "xc = yc")
apply (simp add: same_pnt_zero_rs)

apply (insert the_rest_nat)
apply (drule_tac x="xc" in spec)
apply (drule_tac x="yc" in spec)
apply (simp)
apply (erule exE)
apply (simp)

apply (insert the_rest_nat)
apply (drule_tac x="pair_rev xc" in spec)
apply (drule_tac x="pair_rev yc" in spec)
apply (simp add: pair_rev_neq)
apply (erule exE)
apply (simp)
apply (rename_tac xc yc m n)

apply (simp add: distance_nat_check_def)
apply (simp add: pair_restriction_def)
apply (simp add: pair_rev_def)

apply (insert unique_max_nat)
apply (drule_tac x="(%n. fst xc rest n = fst yc rest n &
                         snd xc rest n = snd yc rest n)" in spec)
apply (drule_tac x="m" in spec)
apply (drule_tac x="n" in spec)
apply (simp add: conj_commute)   (* P & Q = Q & P *)
done

(*****************************************
          pair distance --> max
 *****************************************)

(*** xc ~= yc , snd <= fst ***)

lemma pair_distance_fst_le_neq: 
  "ALL (xc::('a::rs * 'b::rs)) (yc::('a::rs * 'b::rs)).
      (distance(snd xc, snd yc) <= distance(fst xc, fst yc)
       & xc ~= yc & fst xc ~= fst yc)
        --> distance(xc, yc) = max (distance(fst xc, fst yc))
                                   (distance(snd xc, snd yc))"
apply (intro allI impI)
apply (simp add: max_y_x)

apply (insert the_rest_nat)
apply (drule_tac x="xc" in spec)
apply (drule_tac x="yc" in spec)
apply (simp)
apply (erule exE)
apply (simp)

apply (insert the_rest_nat)
apply (drule_tac x="fst xc" in spec)
apply (drule_tac x="fst yc" in spec)
apply (simp)
apply (erule exE)
apply (simp)
apply (rename_tac xc yc m n)

apply (rule order_antisym)

(* n <= m *)

apply (insert distance_nat_check_except_max)
apply (drule_tac x="fst xc" in spec)
apply (drule_tac x="fst yc" in spec)
apply (drule_tac x="n" in spec)
apply (simp)

apply (insert rest_nat_le_1_if)
apply (drule_tac x="snd xc" in spec)
apply (drule_tac x="snd yc" in spec)
apply (drule_tac x="n" in spec)
apply (simp)

apply (insert distance_nat_check_only_max)
apply (drule_tac x="xc" in spec)
apply (drule_tac x="yc" in spec)
apply (drule_tac x="m" in spec)
apply (simp)
apply (drule_tac x="n" in spec)

apply (simp add: pair_restriction_def ALL_realpow_le_le)

(* m <= n *)

apply (drule_tac x="xc" in spec)
apply (rotate_tac -1)
apply (drule_tac x="yc" in spec)
apply (rotate_tac -1)
apply (drule_tac x="m" in spec)

apply (rotate_tac -2)
apply (drule_tac x="fst xc" in spec)
apply (rotate_tac -1)
apply (drule_tac x="fst yc" in spec)
apply (rotate_tac -1)
apply (drule_tac x="n" in spec)
apply (simp)
apply (rotate_tac -1)
apply (drule_tac x="m" in spec)

apply (simp add: pair_restriction_def ALL_realpow_le_le)
done

(*** xc ~= yc , fst <= snd ***)
(*** symmetric with pair_distance_fst_le_neq ***)

lemma pair_distance_snd_le_neq: 
  "ALL (xc::('a::rs * 'b::rs)) (yc::('a::rs * 'b::rs)).
      (distance(fst xc, fst yc) <= distance(snd xc, snd yc)
       & xc ~= yc & snd xc ~= snd yc)
        --> distance(xc, yc) = max (distance(fst xc, fst yc))
                                   (distance(snd xc, snd yc))"
apply (intro allI impI)
apply (insert pair_distance_fst_le_neq)
apply (drule_tac x="pair_rev xc" in spec)
apply (drule_tac x="pair_rev yc" in spec)
apply (simp add: pair_rev_distance_eq)
apply (simp add: pair_rev_def)
apply (simp add: max_commute)
done

(*** xc ~= yc , fst -->snd ***)

lemma pair_distance_fst_snd_le_neq: 
  "ALL (xc::('a::rs * 'b::rs)) (yc::('a::rs * 'b::rs)).
      (distance(fst xc, fst yc) <= distance(snd  xc, snd  yc)
       & fst xc ~= fst yc)
        --> snd xc ~= snd yc"
apply (intro allI impI)
apply (insert diff_pnt_pos_rs)
apply (drule_tac x="fst xc" in spec)
apply (drule_tac x="fst yc" in spec)
apply (simp)
apply (simp add: diff_pnt_pos_rs_inv)
done


(*** xc ~= yc , fst ***)

lemma pair_distance_fst_neq: 
  "ALL (xc::('a::rs * 'b::rs)) (yc::('a::rs * 'b::rs)).
       (xc ~= yc & fst xc ~= fst yc)
        --> distance(xc, yc) = max (distance(fst  xc, fst  yc))
                                   (distance(snd xc, snd yc))"
apply (intro allI impI)

apply (case_tac "distance(snd xc, snd yc) <= distance(fst  xc, fst  yc)")
apply (simp add: pair_distance_fst_le_neq)
apply (simp add: pair_distance_fst_snd_le_neq pair_distance_snd_le_neq)
done

(*** xc ~= yc , snd ***)

lemma pair_distance_snd_neq: 
  "ALL (xc::('a::rs * 'b::rs)) (yc::('a::rs * 'b::rs)).
       (xc ~= yc & snd xc ~= snd yc)
        --> distance(xc, yc) = max (distance(fst xc, fst yc))
                                   (distance(snd xc, snd yc))"

apply (intro allI impI)
apply (insert pair_distance_fst_neq)
apply (drule_tac x="pair_rev xc" in spec)
apply (drule_tac x="pair_rev yc" in spec)
apply (simp add: pair_rev_distance_eq)
apply (simp add: pair_rev_def)
apply (simp add: max_commute)
done

(*** pair_distance --> max distances ***)

lemma pair_distance: 
  "ALL (xc::('a::rs * 'b::rs)) (yc::('a::rs * 'b::rs)).
        distance(xc, yc) = max (distance(fst xc, fst yc))
                               (distance(snd xc, snd yc))"
apply (intro allI)

apply (case_tac "xc = yc")
apply (simp add: same_pnt_zero_rs)

apply (insert neq_pair_decompo)
apply (drule_tac x="xc" in spec)
apply (drule_tac x="yc" in spec)
apply (simp del: disj_not1)       (* del: (~P | Q) = (P --> Q) *)
apply (erule disjE)

apply (simp add: pair_distance_fst_neq)
apply (simp add: pair_distance_snd_neq)
done

(***  max distances --> pair_distance ***)

lemma pair_distance_inv: 
  "ALL (xc::('a::rs * 'b::rs)) (yc::('a::rs * 'b::rs)).
     max (distance(fst  xc, fst  yc)) (distance(snd xc, snd yc))
     = distance(xc, yc)"
apply (simp add: pair_distance)
done

(*** component style ***)

lemma pair_distance_co: 
  "ALL (x1::'a::rs) (x2::'b::rs) (y1::'a::rs) (y2::'b::rs).
        distance((x1, x2), (y1, y2)) 
            = max (distance(x1,y1)) (distance(x2,y2))"
apply (simp add: pair_distance)
done

(*****************************************
               pair maps
 *****************************************)

(*** non_expanding ***)

lemma pair_non_expand:
  "ALL (fc::('a::rs => ('b::rs * 'c::rs))).
     (non_expanding (fst_fun fc) & non_expanding (snd_fun fc))
         --> non_expanding fc"
apply (simp add: non_expanding_def map_alpha_def pair_distance is_lub_def 
                 fst_fun_def snd_fun_def)
done

(*** component style ***)

lemma pair_non_expand_co:
  "ALL (f::('a::rs => 'b::rs)) (g::('a::rs => 'c::rs)).
     (non_expanding f & non_expanding g)
         --> non_expanding (f ** g)"
apply (simp add: non_expanding_def map_alpha_def pair_distance is_lub_def 
                 fst_fun_def snd_fun_def pair_fun_def)
done

(*** alpha_contraction ***)

lemma pair_alpha_contraction:
  "ALL (fc::('a::rs => ('b::rs * 'c::rs))) alpha1 alpha2.
     (contraction_alpha (fst_fun fc) alpha1 &
      contraction_alpha (snd_fun fc) alpha2)
         --> contraction_alpha fc (max alpha1 alpha2)"
apply (intro allI impI)
apply (simp add: contraction_alpha_def map_alpha_def)
apply (simp add: le_max_iff_disj pair_distance)
apply (simp add: fst_fun_def snd_fun_def)
apply (intro allI)
apply (erule conjE)+
apply (drule_tac x="x" in spec)
apply (drule_tac x="x" in spec)
apply (drule_tac x="y" in spec)
apply (drule_tac x="y" in spec)
apply (rule conjI)

apply (insert ALL_real_mult_right_mono)
apply (drule_tac x="alpha1" in spec)
apply (drule_tac x="max alpha1 alpha2" in spec)
apply (drule_tac x="distance (x, y)" in spec)
apply (simp add: positive_rs x_max_x_y)

apply (drule_tac x="alpha2" in spec)
apply (drule_tac x="max alpha1 alpha2" in spec)
apply (drule_tac x="distance (x, y)" in spec)
apply (simp add: positive_rs y_max_x_y)
done

(*** component style ***)

lemma pair_alpha_contraction_co:
  "ALL (f::('a::rs => 'b::rs)) (g::('a::rs => 'c::rs)) alpha1 alpha2.
     (contraction_alpha f alpha1 &
      contraction_alpha g alpha2)
         --> contraction_alpha (f ** g) (max alpha1 alpha2)"
apply (intro allI impI)
apply (insert pair_alpha_contraction)
apply (drule_tac x="f ** g" in spec)
apply (drule_tac x="alpha1" in spec)
apply (drule_tac x="alpha2" in spec)

apply (simp add: pair_fun_def fst_fun_def snd_fun_def)
done

(*** contraction ***)

lemma pair_contraction:
  "ALL (fc::('a::rs => ('b::rs * 'c::rs))).
     (contraction (fst_fun fc) & contraction (snd_fun fc))
         --> contraction fc"
apply (intro allI impI)
apply (simp add: contraction_def)
apply (erule conjE)+
apply (erule exE)+
apply (rule_tac x="max alpha alphaa" in exI)
apply (simp add: pair_alpha_contraction)
done

(*** component style ***)

lemma pair_contraction_co:
  "ALL (f::('a::rs => 'b::rs)) (g::('a::rs => 'c::rs)).
     (contraction f & contraction g) --> contraction (f ** g)"
apply (intro allI)
apply (insert pair_contraction)
apply (drule_tac x="f ** g" in spec)
apply (simp add: pair_fun_def fst_fun_def snd_fun_def)
done

(************************************************************
                       Lemmas (limit)
 ************************************************************)

(*****************************************
                  cauchy
 *****************************************)

(*** cauchy pair_rev = cauchy pair ***)

lemma pair_rev_cauchy_seq:
   "ALL (xcs::(('a::rs * 'b::rs) infinite_seq)).
      cauchy (pair_fun_rev xcs) = cauchy xcs"
apply (simp add: cauchy_def)
apply (simp add: pair_fun_rev_def)
apply (simp add: pair_rev_distance_eq)
done


(*** cauchy pair --> cauchy fst ***)

lemma pair_cauchy_seq_fst:
  "ALL (xcs::(('a::rs * 'b::rs) infinite_seq)).
     cauchy xcs --> cauchy (fst_fun xcs)"
apply (intro allI impI)
apply (simp add: cauchy_def)
apply (intro allI impI)

apply (drule_tac x="delta" in spec)
apply (simp)
apply (erule exE)
apply (rule_tac x="n" in exI)
apply (intro allI impI)
apply (rename_tac xps delta N n m)
apply (drule_tac x="n" in spec)
apply (drule_tac x="m" in spec)
apply (simp add: fst_fun_def)
apply (simp add: pair_distance)
done

(*** cauchy pair --> cauchy snd ***)

lemma pair_cauchy_seq_snd:
  "ALL (xcs::(('a::rs * 'b::rs) infinite_seq)).
     cauchy xcs --> cauchy (snd_fun xcs)"
apply (intro allI impI)
apply (insert pair_cauchy_seq_fst)
apply (drule_tac x="pair_fun_rev xcs" in spec)
apply (simp add: pair_rev_cauchy_seq)
apply (simp add: pair_fst_fun_rev)
done

(*** cauchy pair --> cauchy both ***)

lemma pair_cauchy_seq:
  "ALL (xcs::(('a::rs * 'b::rs) infinite_seq)).
     cauchy xcs --> (cauchy (fst_fun xcs) & cauchy (snd_fun xcs))"
apply (simp add: pair_cauchy_seq_fst pair_cauchy_seq_snd)
done

(*** component style ***)

lemma pair_cauchy_seq_co:
  "ALL (xs::('a::rs infinite_seq)) (ys::('b::rs infinite_seq)).
     cauchy (xs ** ys) --> (cauchy xs & cauchy ys)"
apply (intro allI impI)
apply (insert pair_cauchy_seq)
apply (drule_tac x="xs ** ys" in spec)
apply (simp add: fst_fun_def snd_fun_def pair_fun_def)
done

(*****************************************
                  normal
 *****************************************)

(*** normal pair_rev = normal pair ***)

lemma pair_rev_normal_seq:
   "ALL (xcs::(('a::rs * 'b::rs) infinite_seq)).
      normal (pair_fun_rev xcs) = normal xcs"
apply (simp add: normal_def)
apply (simp add: pair_fun_rev_def pair_rev_distance_eq)
done

(*** normal pair --> normal fst ***)

lemma pair_normal_seq_fst:
  "ALL (xcs::(('a::rs * 'b::rs) infinite_seq)).
     normal xcs --> normal (fst_fun xcs)"
apply (intro allI impI)
apply (simp add: normal_def)
apply (intro allI)

apply (drule_tac x="n" in spec)
apply (drule_tac x="m" in spec)
apply (simp add: fst_fun_def)
apply (simp add: pair_distance)
done

(*** normal pair --> normal snd ***)

lemma pair_normal_seq_snd:
  "ALL (xcs::(('a::rs * 'b::rs) infinite_seq)).
     normal xcs --> normal (snd_fun xcs)"
apply (intro allI impI)
apply (insert pair_normal_seq_fst)
apply (drule_tac x="pair_fun_rev xcs" in spec)
apply (simp add: pair_rev_normal_seq)
apply (simp add: pair_fst_fun_rev)
done

(*** normal pair --> normal both ***)

lemma pair_normal_seq:
  "ALL (xcs::(('a::rs * 'b::rs) infinite_seq)).
     normal xcs --> (normal (fst_fun xcs) & normal (snd_fun xcs))"
apply (simp add: pair_normal_seq_fst pair_normal_seq_snd)
done

(*** component style ***)

lemma pair_normal_seq_co:
  "ALL (xs::('a::rs infinite_seq)) (ys::('b::rs infinite_seq)).
     normal (xs ** ys) --> (normal xs & normal ys)"
apply (intro allI impI)
apply (insert pair_normal_seq)
apply (drule_tac x="xs ** ys" in spec)
apply (simp add: fst_fun_def snd_fun_def)
apply (simp add: pair_fun_def)
done

(*****************************************
                 limit
 *****************************************)

(*** limit pair_rev = limit pair ***)

lemma pair_rev_limit:
   "ALL (xcs::(('a::rs * 'b::rs) infinite_seq)) yc.
       (pair_fun_rev xcs) convergeTo (pair_rev yc) = xcs convergeTo yc"
apply (simp add: convergeTo_def)
apply (simp add: pair_fun_rev_def pair_rev_distance_eq)
done


(*** limit (single <-- pair) ***)

(*** fst ***)

lemma pair_limit_from_fst:
  "ALL (xcs::(('a::rs * 'b::rs) infinite_seq)) yc.
     xcs convergeTo yc --> (fst_fun xcs) convergeTo fst yc"
apply (intro allI impI)
apply (simp add: convergeTo_def)
apply (intro allI impI)

apply (drule_tac x="eps" in spec)
apply (simp)
apply (erule exE)
apply (rule_tac x="n" in exI)
apply (intro allI impI)
apply (rename_tac xcs yc eps N n)
apply (drule_tac x="n" in spec)
apply (simp add: fst_fun_def)
apply (simp add: pair_distance)
done

(*** snd ***)

lemma pair_limit_from_snd:
  "ALL (xcs::(('a::rs * 'b::rs) infinite_seq)) yc.
     xcs convergeTo yc --> (snd_fun xcs) convergeTo snd yc"
apply (intro allI impI)
apply (insert pair_limit_from_fst)
apply (drule_tac x="pair_fun_rev xcs" in spec)
apply (drule_tac x="pair_rev yc" in spec)
apply (simp add: pair_rev_limit)
apply (simp add: pair_fst_fun_rev pair_snd_fun_rev pair_rev_def)
done

(*** limit (single --> pair) (note: rs) ***)

lemma pair_limit_to:
  "ALL (xcs::(('a::rs * 'b::rs) infinite_seq)) y z.
      (fst_fun xcs convergeTo y &
       snd_fun xcs convergeTo z)
          --> xcs convergeTo (y, z)"
apply (intro allI impI)
apply (simp add: convergeTo_def)
apply (intro allI impI)

apply (erule conjE)
apply (drule_tac x="eps" in spec)
apply (drule_tac x="eps" in spec)
apply (simp)
apply (erule exE)+
apply (rename_tac xcs y z eps N2 N1)

apply (rule_tac x="max N1 N2" in exI)
apply (intro allI impI)
apply (drule_tac x="m" in spec)

apply (simp add: pair_distance)
apply (simp add: fst_fun_def snd_fun_def)
done

(*****************************************
     limit (cms and cauchy --> limit) 
 *****************************************)

lemma pair_cms_cauchy_limit:
  "ALL (xcs::(('a::cms_rs * 'b::cms_rs) infinite_seq)).
      cauchy xcs --> xcs convergeTo (pair_limit xcs)"
apply (intro allI impI)
apply (insert pair_limit_to)
apply (drule_tac x="xcs" in spec)
apply (drule_tac x="limit (fst_fun xcs)" in spec)
apply (drule_tac x="limit (snd_fun xcs)" in spec)

apply (simp add: the_limit pair_cauchy_seq)
apply (simp add: pair_limit_def)
done

(*****************************************
     Pair Complete Metric Space (RS)
 *****************************************)

lemma pair_cms_rs:
  "ALL (xcs::(('a::cms_rs * 'b::cms_rs) infinite_seq)).
       cauchy xcs --> (EX yc. xcs convergeTo yc)"
apply (intro allI impI)
apply (rule_tac x="pair_limit xcs" in exI)
apply (simp add: pair_cms_cauchy_limit)
done


(************************************************************
                   Pair RS ==> CMS and RS
 ************************************************************)

instance * :: (cms_rs,cms_rs) cms
apply (intro_classes)
apply (insert pair_cms_rs)
apply (simp)
done

instance * :: (cms_rs,cms_rs) cms_rs
apply (intro_classes)
done

(*----------------------------------------------------------*
 |                                                          |
 |                Continuity on Metric space                |
 |                                                          |
 *----------------------------------------------------------*)

declare disj_not1 [simp del]

lemma pair_continuous: 
   "ALL R1 R2. continuous R1 & continuous R2 
          --> continuous (%xc. (R1 (fst xc) & R2 (snd xc)))"
apply (intro allI impI)
apply (simp add: continuous_def)
apply (intro allI)
apply (rule conjI)

apply (intro impI)
apply (erule conjE)
apply (drule_tac x="a" in spec)
apply (simp)
apply (erule exE)
apply (rule_tac x="n" in exI)
apply (intro allI impI)
apply (rotate_tac -1)
apply (drule_tac x="b" in spec)
apply (simp add: pair_restriction_def)

apply (intro impI)
apply (erule conjE)
apply (rotate_tac -1)
apply (drule_tac x="b" in spec)
apply (simp)
apply (erule exE)
apply (rule_tac x="n" in exI)
apply (intro allI impI)
apply (rotate_tac -2)
apply (drule_tac x="ba" in spec)
apply (simp add: pair_restriction_def)
done

declare disj_not1 [simp]

end
